import os
import requests
from autotrader.brokers.trading import Order
from autotrader.comms.notifier import Notifier
from autotrader.utilities import read_yaml, write_yaml, get_logger


class Telegram(Notifier):
    """Simple telegram bot to send messages.

    To use this, you must first create a Telegram bot via the BotFather. Then,
    provide the API token generated here as the api_token. If you do not know
    your chat_id, send the bot a message on telegram, and it will be inferred
    when this class is instantiated with the api_token.
    """

    def __repr__(self) -> str:
        return "AutoTrader-Telegram communication module"

    def __init__(
        self, api_token: str, chat_id: str = None, logger_kwargs: dict = None
    ) -> None:
        """Instantiate the bot.

        Parameters
        ----------
        token : str
            The bot API token.

        chat_id : str, optional
            The default chat_id to send messages to.
        """
        # Create logger
        logger_kwargs = logger_kwargs if logger_kwargs else {}
        self.logger = get_logger(name="telegram_combot", **logger_kwargs)

        # Save attributes
        self.token = api_token
        if chat_id is None:
            # Try get chat ID
            self.logger.info(
                "No chat ID specified - attempting to load from recent updates."
            )
            _, chat_id = self.get_chat_id()
        self.chat_id = chat_id

    def get_chat_id(self):
        response = requests.get(f"https://api.telegram.org/bot{self.token}/getUpdates")
        try:
            chat = response.json()["result"][-1]["message"]["chat"]
            chat_id = chat["id"]
            name = chat["first_name"]
            self.logger.info(f"Found chat ID for {name}: {chat_id}.")

            # Write ID to file for future
            path = "config/keys.yaml"
            if os.path.exists(path):
                # Config file exists, proceed
                config = read_yaml(path)

                if "TELEGRAM" in config:
                    # Telegram in config
                    self.logger.info("Adding chat_id to configuration file.")
                    if "chat_id" not in config["TELEGRAM"]:
                        # Add chat ID
                        config["TELEGRAM"]["chat_id"] = chat_id
                else:
                    # Telegram not in config; insert fresh
                    self.logger.info(
                        "Adding telegram configuration details to configuration file."
                    )
                    config["TELEGRAM"] = {
                        "api_key": self.token,
                        "chat_id": chat_id,
                    }

                # Write to file
                write_yaml(config, path)

            return name, chat_id

        except IndexError:
            # No updates to read from
            self.logger.error(
                "Cannot find chat ID - please make sure you have recently messaged the bot."
            )
            return None, None

    def send_message(self, message: str, chat_id: str = None, *args, **kwargs):
        if chat_id is None:
            chat_id = self.chat_id
        self.logger.debug(f"Sending message to {chat_id}: {message}")
        url_req = f"https://api.telegram.org/bot{self.token}/sendMessage?chat_id={chat_id}&text={message}"
        response = requests.get(url_req)
        if response.status_code != 200:
            self.logger.error(f"Failed to send message to {chat_id}: {response.reason}")

    def send_order(self, order: Order, *args, **kwargs) -> None:
        side = "long" if order.direction > 0 else "short"
        message = (
            f"New {order.instrument} {order.order_type} order created: "
            + f"{order.size} units {side}"
        )

        # Create bot and send message
        self.send_message(message)



import os
import importlib
import traceback
from datetime import datetime
from autotrader.strategy import Strategy
from autotrader.brokers.trading import Order
from autotrader.brokers.broker import Broker
from typing import TYPE_CHECKING, Literal, Union
from autotrader.utilities import TradeAnalysis, get_logger
from autotrader.brokers.virtual import Broker as VirtualBroker

if TYPE_CHECKING:
    from autotrader import AutoTrader
    from autotrader.comms.notifier import Notifier


class AutoTraderBot:
    """AutoTrader Trading Bot, responsible for a trading strategy."""

    def __init__(
        self,
        instrument: str,
        strategy_dict: dict,
        broker: Union[dict[str, Broker], Broker],
        autotrader_instance: "AutoTrader",
    ) -> None:
        """Instantiates an AutoTrader Bot.

        Parameters
        ----------
        instrument : str
            The trading instrument assigned to the bot instance.

        strategy_dict : dict
            The strategy configuration dictionary.

        broker : AutoTrader Broker instance
            The AutoTrader Broker module.

        autotrader_instance : AutoTrader
            The parent AutoTrader instance.

        Raises
        ------
        Exception
            When there is an error retrieving the instrument data.
        """
        # Type hint inherited attributes
        self._multiple_brokers: bool
        self._global_config_dict: dict[str, any]
        self._environment: Literal["paper", "live"]
        self._broker_name: str
        self._feed: str
        self._base_currency: str
        self._data_path_mapper: dict
        self._backtest_mode: bool
        self._data_directory: str
        self._papertrading: bool
        self._scan_mode: bool
        self._notify: int
        self._verbosity: int
        self._notifier: "Notifier"
        self._virtual_tradeable_instruments: dict[str, list[str]]
        self._logger_kwargs: dict
        self._data_start: datetime
        self._data_end: datetime
        self._dynamic_data: bool

        # Inherit user options from autotrader
        for attribute, value in autotrader_instance.__dict__.items():
            setattr(self, attribute, value)

        # Create autobot logger
        self.logger = get_logger(name="autobot", **self._logger_kwargs)

        # Assign local attributes
        self.instrument = instrument
        self._broker: Broker = broker

        # Define execution framework
        if self._execution_method is None:
            self._execution_method = self._submit_order

        # Check for muliple brokers and construct mapper
        if self._multiple_brokers:
            # Trading across multiple venues
            self._brokers: list[Broker] = self._broker
            self._instrument_to_broker = {}
            for (
                broker_name,
                tradeable_instruments,
            ) in self._virtual_tradeable_instruments.items():
                for instrument in tradeable_instruments:
                    if instrument in self._instrument_to_broker:
                        # Instrument is already in mapper, add broker
                        self._instrument_to_broker[instrument].append(
                            self._brokers[broker_name]
                        )
                    else:
                        # New instrument, add broker
                        self._instrument_to_broker[instrument] = [
                            self._brokers[broker_name]
                        ]

        else:
            # Trading through a single broker
            self._brokers = {self._broker_name: self._broker}

            # Map instruments to broker
            self._instrument_to_broker = {}
            instruments = [instrument] if isinstance(instrument, str) else instrument
            for instrument in instruments:
                self._instrument_to_broker[instrument] = [self._broker]

        # Unpack strategy parameters and assign to strategy_params
        strategy_config = strategy_dict["config"]
        interval = strategy_config["INTERVAL"]
        risk_pc = strategy_config["RISK_PC"] if "RISK_PC" in strategy_config else None
        sizing = strategy_config["SIZING"] if "SIZING" in strategy_config else None
        params = (
            strategy_config["PARAMETERS"] if "PARAMETERS" in strategy_config else {}
        )
        strategy_params = params
        strategy_params["granularity"] = (
            strategy_params["granularity"]
            if "granularity" in strategy_params
            else interval
        )
        strategy_params["risk_pc"] = (
            strategy_params["risk_pc"] if "risk_pc" in strategy_params else risk_pc
        )
        strategy_params["sizing"] = (
            strategy_params["sizing"] if "sizing" in strategy_params else sizing
        )
        self._strategy_params = strategy_params

        # Import Strategy
        if strategy_dict["class"] is not None:
            strategy = strategy_dict["class"]
        else:
            strat_module = strategy_config["MODULE"]
            strat_name = strategy_config["CLASS"]
            strat_package_path = os.path.join(self._home_dir, "strategies")
            strat_module_path = os.path.join(strat_package_path, strat_module) + ".py"
            strat_spec = importlib.util.spec_from_file_location(
                strat_module, strat_module_path
            )
            strategy_module = importlib.util.module_from_spec(strat_spec)
            strat_spec.loader.exec_module(strategy_module)
            strategy = getattr(strategy_module, strat_name)

        # Strategy shutdown routine
        self._strategy_shutdown_method = strategy_dict["shutdown_method"]

        # Check for portfolio strategy
        trade_portfolio = (
            strategy_config["PORTFOLIO"] if "PORTFOLIO" in strategy_config else False
        )

        portfolio = strategy_config["WATCHLIST"] if trade_portfolio else False

        # Initialise the broker with a cache of data, if backtesting over
        # full dataset, or otherwise just the window specified in the strategy.
        # Initialise broker datasets
        for instrument, brokers in self._instrument_to_broker.items():
            for broker in brokers:
                broker._initialise_data(
                    **{
                        "instrument": instrument,
                        "data_start": self._data_start,
                        "data_end": self._data_end,
                        "granularity": interval,
                    }
                )

        # Build strategy instantiation arguments
        strategy_inputs = {
            "parameters": params,
            "instrument": self.instrument,
            "broker": self._broker,
            "notifier": self._notifier,
            "logger_kwargs": self._logger_kwargs,
        }

        # Instantiate Strategy
        my_strat: Strategy = strategy(**strategy_inputs)

        # Assign strategy to local attributes
        self._last_bars = None
        self._strategy = my_strat
        self._strategy_name = (
            strategy_config["NAME"]
            if "NAME" in strategy_config
            else "(unnamed strategy)"
        )

        # Assign stop trading method to strategy
        self._strategy.stop_trading = autotrader_instance._remove_instance_file

        # Call indicators plotting method
        if self._backtest_mode:
            data = self._broker._data_cache[self.instrument]
            my_strat.create_plotting_indicators(data)

        # Assign strategy attributes for tick-based strategy development
        # TODO - improve type hints using strategy base class
        if self._backtest_mode:
            self._strategy._backtesting = True
            self.trade_results = None
        if interval.split(",")[0] == "tick":
            self._strategy._tick_data = True

    def __repr__(self):
        if isinstance(self.instrument, list):
            return "Portfolio AutoTraderBot"
        else:
            return f"{self.instrument} AutoTraderBot"

    def __str__(self):
        return "AutoTraderBot instance"

    def _update(self, timestamp: datetime) -> None:
        """Update strategy with the latest data and generate a trade signal.

        Parameters
        ----------
        timestamp : datetime, optional
            The current update time.
        """
        if self._backtest_mode or self._papertrading:
            # Update virtual broker
            self._update_virtual_broker(dt=timestamp)

        # Call strategy for orders
        try:
            strategy_orders = self._strategy.generate_signal(timestamp)
        except Exception as e:
            self.logger.error(f"Error when updating strategy: {e}")
            self.logger.info(traceback.format_exc())
            strategy_orders = []

        # Check and qualify orders
        orders = self._check_orders(strategy_orders)
        self._qualify_orders(orders)

        if not self._scan_mode:
            # Submit orders
            for order in orders:
                # Submit order to relevant exchange
                try:
                    self._execution_method(
                        broker=self._brokers[order.exchange],
                        order=order,
                        order_time=timestamp,
                    )
                except Exception as e:
                    traceback_str = "".join(traceback.format_tb(e.__traceback__))
                    exception_str = f"AutoTrader exception when submitting order: {e}"
                    print_str = exception_str + "\nTraceback:\n" + traceback_str
                    self.logger.error(print_str)

        # If paper trading, update virtual broker again to trigger any orders
        if self._papertrading:
            self._update_virtual_broker(dt=timestamp)

        # Log message
        current_time = timestamp.strftime("%b %d %Y %H:%M:%S")
        if len(orders) > 0:
            for order in orders:
                direction = "long" if order.direction > 0 else "short"
                order_string = (
                    f"{current_time}: {order.instrument} "
                    + f"{direction} {order.order_type} order of "
                    + f"{order.size} units placed."
                )
                self.logger.info(order_string)
        else:
            self.logger.debug(
                f"{current_time}: No signal detected ({self.instrument})."
            )

        # Check for orders placed and/or scan hits
        if int(self._notify) > 0 and not (self._backtest_mode or self._scan_mode):
            # TODO - what is this conditional?
            for order in orders:
                self._notifier.send_order(order)

        # Check scan results
        if self._scan_mode:
            # Report AutoScan results
            if int(self._verbosity) > 0 or int(self._notify) == 0:
                # Scan reporting with no notifications requested
                if len(orders) == 0:
                    print(f"{self.instrument}: No signal detected.")

                else:
                    # Scan detected hits
                    print("Scan hits:")
                    for order in orders:
                        print(order)

            if int(self._notify) > 0:
                # Notifications requested
                for order in orders:
                    self._notifier.send_message(f"Scan hit: {order}")

    def _check_orders(self, orders) -> list[Order]:
        """Checks that orders returned from strategy are in the correct
        format.

        Returns
        -------
        List of Orders

        Notes
        -----
        An order must have (at the very least) an order type specified. Usually,
        the direction will also be required, except in the case of close order
        types. If an order with no order type is provided, it will be ignored.
        """

        def check_type(orders):
            checked_orders = []
            if isinstance(orders, dict):
                # Order(s) provided in dictionary
                if "order_type" in orders:
                    # Single order dict provided
                    if "instrument" not in orders:
                        orders["instrument"] = self.instrument
                    checked_orders.append(Order._from_dict(orders))

                elif len(orders) > 0:
                    # Multiple orders provided
                    for key, item in orders.items():
                        if isinstance(item, dict) and "order_type" in item:
                            # Convert order dict to Order object
                            if "instrument" not in item:
                                item["instrument"] = self.instrument
                            checked_orders.append(Order._from_dict(item))
                        elif isinstance(item, Order):
                            # Native Order object, append as is
                            checked_orders.append(item)
                        else:
                            raise Exception(f"Invalid order submitted: {item}")

                elif len(orders) == 0:
                    # Empty order dict
                    pass

            elif isinstance(orders, Order):
                # Order object directly returned
                checked_orders.append(orders)

            elif isinstance(orders, list):
                # Order(s) provided in list
                for item in orders:
                    if isinstance(item, dict) and "order_type" in item:
                        # Convert order dict to Order object
                        if "instrument" not in item:
                            item["instrument"] = self.instrument
                        checked_orders.append(Order._from_dict(item))
                    elif isinstance(item, Order):
                        # Native Order object, append as is
                        checked_orders.append(item)
                    else:
                        raise Exception(f"Invalid order submitted: {item}")
            else:
                raise Exception(f"Invalid order/s submitted: '{orders}' received")

            return checked_orders

        def add_strategy_data(orders: list[Order]):
            # Append strategy parameters to each order
            for order in orders:
                order.instrument = (
                    self.instrument if not order.instrument else order.instrument
                )
                order.strategy = (
                    self._strategy.name
                    if "name" in self._strategy.__dict__
                    else self._strategy_name
                )
                order.granularity = self._strategy_params["granularity"]
                order._sizing = self._strategy_params["sizing"]
                order._risk_pc = self._strategy_params["risk_pc"]

        def check_order_details(orders: list[Order]) -> None:
            # Check details for order type have been provided
            for ix, order in enumerate(orders):
                order.instrument = (
                    order.instrument
                    if order.instrument is not None
                    else self.instrument
                )
                if order.order_type in ["market", "limit", "stop-limit", "reduce"]:
                    if not order.direction:
                        # Order direction was not provided, delete order
                        del orders[ix]
                        continue

                # Check that an exchange has been specified
                if order.exchange is None:
                    # Exchange not specified
                    if self._multiple_brokers:
                        # Trading across multiple venues
                        raise Exception(
                            "The exchange to which an order is to be "
                            + "submitted must be specified when trading across "
                            + "multiple venues. Please include the 'exchange' "
                            + "argument when creating an order."
                        )
                    else:
                        # Trading on single venue, auto fill
                        order.exchange = self._broker_name

                if order.size is None and order.order_type != "modify":
                    raise Exception("Must provide order size.")

        # Perform checks
        if orders is not None:
            checked_orders = check_type(orders)
            add_strategy_data(checked_orders)
            check_order_details(checked_orders)
            return checked_orders
        else:
            # Return empty list
            return []

    def _qualify_orders(self, orders: list[Order]) -> None:
        """Prepare orders for submission."""
        for order in orders:
            # Get relevant broker
            broker: Broker = self._brokers[order.exchange]

            # Determine current price to assign to order
            if self._feed != "none":
                # Get order price from current orderbook
                orderbook = broker.get_orderbook(
                    instrument=order.instrument,
                )

                # Check order type to assign variables
                if order.direction < 0:
                    order_price = orderbook.bids.loc[0]["price"]
                else:
                    order_price = orderbook.asks.loc[0]["price"]

            else:
                # Do not provide order price yet
                order_price = None

            # Call order to update
            order(order_price=order_price)

    def _update_virtual_broker(self, dt: datetime) -> None:
        """Updates the virtual broker state. Only called when backtesting or paper trading."""
        for instrument, brokers in self._instrument_to_broker.items():
            for broker in brokers:
                broker: VirtualBroker
                broker._update_positions(instrument=instrument, dt=dt)

    def _create_trade_results(self, broker_histories: dict) -> dict:
        """Constructs bot-specific trade summary for post-processing."""
        trade_results = TradeAnalysis(self._broker, broker_histories, self.instrument)
        trade_results.indicators = (
            self._strategy.indicators if hasattr(self._strategy, "indicators") else None
        )
        trade_results.data = self._broker._data_cache[self.instrument]
        trade_results.interval = self._strategy_params["granularity"]
        self.trade_results = trade_results

    def _check_strategy_for_plot_data(self, use_strat_plot_data: bool = False):
        """Checks the bot's strategy to see if it has the plot_data attribute.

        Returns
        -------
        plot_data : pd.DataFrame
            The data to plot.

        Notes
        -----
        This method is a placeholder for a future feature, allowing
        customisation of what is plotted by setting plot_data and plot_type
        attributes from within a strategy.
        """
        strat_params = self._strategy.__dict__
        if "plot_data" in strat_params and use_strat_plot_data:
            plot_data = strat_params["plot_data"]
        else:
            plot_data = self._broker.get_candles(instrument=self.instrument)

        return plot_data

    def _strategy_shutdown(
        self,
    ):
        """Perform the strategy shutdown routines, if they exist."""
        if self._strategy_shutdown_method is not None:
            try:
                shutdown_method = getattr(
                    self._strategy, self._strategy_shutdown_method
                )
                shutdown_method()
            except AttributeError:
                self.logger.error(
                    f"\nShutdown method '{self._strategy_shutdown_method}' not found!"
                )

    @staticmethod
    def _submit_order(broker: Broker, order: Order, *args, **kwargs):
        "The default order execution method."
        broker.place_order(order, *args, **kwargs)



import os
import time
import click
import shutil
import requests
import autotrader


def download_file(url):
    local_filename = url.split("/")[-1]
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(local_filename, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    return local_filename


@click.group()
def cli():
    """AutoTrader command line interface."""
    pass


@click.command()
def version():
    """Shows the installed version number of AutoTrader."""
    print(autotrader.__version__)


@click.command()
@click.option(
    "-s",
    "--strategies",
    help="The name of strategies to include in the initialised directory.",
)
@click.option(
    "-d",
    "--demo",
    is_flag=True,
    show_default=True,
    default=False,
    help="Initialise the directory with the AutoTrader demo repository.",
)
@click.argument("name", default=".")
def init(strategies, demo, name):
    """Initialises the directory NAME for trading with AutoTrader. If no
    directory NAME is provided, the current directory will be initialised.

    To include ready-to-go strategies in the initialised directory,
    specify them using the strategies option. You can provide the following
    arguments:

    \b
    - template: a strategy template module
    - config: a strategy configuration template
    - strategy_name: the name of a strategy to load

    Strategies are loaded from the AutoTrader demo repository here:
    https://github.com/kieran-mackle/autotrader-demo
    """
    autotrader.utilities.print_banner()

    # Construct filepaths
    file_dir = os.path.dirname(os.path.abspath(__file__))
    data_dir = os.path.join(file_dir, "..", "package_data")
    if name != ".":
        # Initialise directory specified
        dir_name = name
        if not os.path.isdir(dir_name):
            # Directory doesn't exist yet - create it
            os.mkdir(dir_name)
    else:
        # Initialise current directory
        dir_name = os.path.abspath(os.getcwd())

    if demo:
        # Clone demo repository
        click.echo("Initialising from demo repository.\n")

        demo_url = "https://github.com/kieran-mackle/autotrader-demo"
        os.system(f"git clone {demo_url} {dir_name}")
        click.echo("\nAutoTrader initialisation complete.")
        return

    # Check if config directory exists
    config_dir = os.path.join(dir_name, "config")
    if not os.path.isdir(config_dir):
        os.mkdir(config_dir)

    # Copy keys config
    keys_config_fp = os.path.join(data_dir, "keys.yaml")
    shutil.copyfile(keys_config_fp, os.path.join(config_dir, "keys.yaml"))

    # Check if strategy directory exists
    strategy_dir = os.path.join(dir_name, "strategies")
    if not os.path.isdir(strategy_dir):
        # Strategy directory doesn't exist - create it
        os.mkdir(strategy_dir)

    # Add strategies
    valid_args = [
        "config",
        "template",
        "macd",
        "ema_crossover",
        "long_ema_crossover",
        "supertrend",
        "rebalance",
    ]
    if strategies is not None:
        for strategy in strategies.replace(" ", "").split(","):
            strategy = strategy.lower()

            # Check
            if strategy not in valid_args:
                raise Exception(f"{strategy} is not a valid argument.")

            # Construct urls
            if strategy == "template":
                # Get strategy template from main repo
                urls = {
                    "strategies": "https://raw.githubusercontent.com/"
                    + "kieran-mackle/AutoTrader/main/templates/strategy.py",
                    "config": None,
                }

            elif strategy == "config":
                # Get strategy config file
                urls = {
                    "strategies": None,
                    "config": "https://raw.githubusercontent.com/"
                    + "kieran-mackle/AutoTrader/main/templates/strategy_config.yaml",
                }

            else:
                # Get from demo repo
                urls = {
                    "strategies": "https://raw.githubusercontent.com/kieran-mackle/"
                    + f"autotrader-demo/main/strategies/{strategy}.py",
                    "config": "https://raw.githubusercontent.com/kieran-mackle/"
                    + f"autotrader-demo/main/config/{strategy}.yaml",
                }

            for dir in ["strategies", "config"]:
                try:
                    # Download
                    filename = download_file(urls[dir])

                    # Move to appropriate directory
                    move_to = os.path.join(dir_name, dir, filename)
                    os.rename(filename, move_to)

                except:
                    pass

    # Print completion message
    click.echo("AutoTrader initialisation complete.")


@click.command()
def demo():
    """Runs a demo backtest in AutoTrader."""
    # Download the strategy file and data
    print("Loading demo files...")
    branch = "main"
    strat_filename = download_file(
        "https://raw.githubusercontent.com/kieran-mackle/"
        + f"AutoTrader/{branch}/tests/macd_strategy.py"
    )
    print("  Done.")

    # Run backtest
    os.system("python3 macd_strategy.py")

    # Clean up files
    os.remove(strat_filename)


@click.command()
@click.option("-p", "--port", default=8009, help="The port to serve data to.")
@click.option(
    "-i",
    "--initial-nav",
    default=None,
    help="The initial NAV to use for relative PnL calculations.",
)
@click.option(
    "-m",
    "--max-nav",
    default=None,
    help="The maximum NAV to use for drawdown calculations.",
)
@click.option(
    "-f", "--picklefile", help="The pickle file containing a virtual broker instance."
)
@click.option("-c", "--config", help="The monitor yaml configuration filepath.")
@click.option("-b", "--broker", help="The name of the broker to connect to.")
@click.option("-e", "--environment", default="paper", help="The trading environment.")
def monitor(port, initial_nav, max_nav, picklefile, config, broker, environment):
    """Monitors a broker/exchange and serves the information
    to a prometheus database.
    """
    # Print banner
    autotrader.utilities.print_banner()

    # Construct config dictionary
    if config is not None:
        # Read from file
        monitor_config = autotrader.utilities.read_yaml(config)

    else:
        monitor_config = {
            "port": port,
            "broker": broker,
            "picklefile": picklefile,
            "initial_nav": initial_nav,
            "max_nav": max_nav,
            "environment": environment,
            "sleep_time": 30,
        }

    # Create Monitor instance
    monitor = autotrader.utilities.Monitor(**monitor_config)

    # Run
    monitor.run()


@click.command()
@click.argument("pickle")
def snapshot(pickle):
    """Prints a snapshot of the trading account of a broker PICKLE instance
    file."""
    autotrader.utilities.print_banner()
    autotrader.AutoTrader.papertrade_snapshot(pickle)
    print("")


# Add commands to CLI
cli.add_command(version)
cli.add_command(init)
cli.add_command(monitor)
cli.add_command(snapshot)
cli.add_command(demo)



import pandas as pd
from datetime import datetime
from abc import ABC, abstractmethod
from autotrader.brokers.trading import Order, OrderBook


class AbstractBroker(ABC):
    @abstractmethod
    def __init__(self, config: dict) -> None:
        """AutoTrader Broker Class constructor."""

    @abstractmethod
    def __repr__(self):
        return "AutoTrader Broker interface"

    @abstractmethod
    def __str__(self):
        return "AutoTrader Broker interface"

    @property
    @abstractmethod
    def data_broker(self) -> "AbstractBroker":
        pass

    @abstractmethod
    def get_NAV(self, *args, **kwargs) -> float:
        """Returns the net asset/liquidation value of the account."""

    @abstractmethod
    def get_balance(self, *args, **kwargs) -> float:
        """Returns account balance."""

    @abstractmethod
    def place_order(self, order: Order, *args, **kwargs) -> None:
        """Translate order and place via exchange API."""

    @abstractmethod
    def get_orders(self, instrument: str = None, *args, **kwargs) -> dict:
        """Returns all pending orders (have not been filled) in the account."""

    @abstractmethod
    def cancel_order(self, order_id: int, *args, **kwargs) -> None:
        """Cancels order by order ID."""

    @abstractmethod
    def get_trades(self, instrument: str = None, *args, **kwargs) -> dict:
        """Returns the trades (fills) made by the account."""

    @abstractmethod
    def get_positions(self, instrument: str = None, *args, **kwargs) -> dict:
        """Gets the current positions open on the account.

        Parameters
        ----------
        instrument : str, optional
            The trading instrument name (symbol). The default is None.

        Returns
        -------
        open_positions : dict
            A dictionary containing details of the open positions.
        """

    @abstractmethod
    def get_candles(
        self,
        instrument: str,
        granularity: str = None,
        count: int = None,
        start_time: datetime = None,
        end_time: datetime = None,
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        """Get the historical OHLCV candles for an instrument."""

    @abstractmethod
    def get_orderbook(self, instrument: str, *args, **kwargs) -> OrderBook:
        """Get the orderbook for an instrument."""

    @abstractmethod
    def get_public_trades(self, instrument: str, *args, **kwargs):
        """Get the public trade history for an instrument."""

    @abstractmethod
    def _initialise_data(self, *args, **kwargs):
        """Initialise the broker data."""

    @abstractmethod
    def get_precision(self, instrument: str):
        """Return the price and amount precision for an instrument."""


class Broker(AbstractBroker):
    def configure(self, *args, **kwargs):
        """Generic configure method, placeholder for typehinting. Only required
        for the virtual broker."""

    def get_NAV(self, *args, **kwargs) -> float:
        raise NotImplementedError

    def get_balance(self, *args, **kwargs) -> float:
        raise NotImplementedError

    def place_order(self, order: Order, *args, **kwargs) -> None:
        raise NotImplementedError

    def get_orders(self, instrument: str = None, *args, **kwargs) -> dict:
        raise NotImplementedError

    def cancel_order(self, order_id: int, *args, **kwargs) -> None:
        raise NotImplementedError

    def get_trades(self, instrument: str = None, *args, **kwargs) -> dict:
        raise NotImplementedError

    def get_positions(self, instrument: str = None, *args, **kwargs) -> dict:
        raise NotImplementedError

    def _initialise_data(self, *args, **kwargs):
        pass

    def get_precision(self, instrument: str):
        return None



import time
import ccxt
import pandas as pd
from datetime import datetime, timezone
from autotrader.utilities import get_logger
from autotrader.brokers.broker import Broker
from autotrader.brokers.trading import OrderBook
from autotrader.brokers.trading import Order, Trade, Position


class Broker(Broker):
    def __init__(self, config: dict) -> None:
        """AutoTrader Broker Class constructor."""
        # Unpack config and connect to broker-side API
        self.exchange: str = config["exchange"]
        exchange_instance = getattr(ccxt, self.exchange)
        if "api_key" in config:
            ccxt_config = {
                "apiKey": config["api_key"],
                "secret": config["secret"],
                "options": config["options"],
                "password": config["password"],
            }
        else:
            ccxt_config = {}

        # Create logger
        self._logger = get_logger(
            name=f"{self.exchange}_broker", **config["logging_options"]
        )

        # Instantiate exchange connection
        self.api: ccxt.Exchange = exchange_instance(ccxt_config)

        # Set sandbox mode
        self._sandbox_str = ""
        if config.get("sandbox_mode", False):
            self.api.set_sandbox_mode(True)
            self._sandbox_str = " (sandbox mode)"

        # Load markets
        self._logger.info(f"Loading instruments for {self.exchange}.")
        markets = self.api.load_markets()
        self.base_currency = config["base_currency"]

        # Assign data broker
        self._data_broker = self

        # Stored instrument precisions
        self._instrument_precisions = {}

    def __repr__(self):
        return (
            f"AutoTrader-{self.exchange[0].upper()}"
            + f"{self.exchange[1:].lower()} interface"
            + self._sandbox_str
        )

    def __str__(self):
        return self.__repr__()

    @property
    def data_broker(self):
        return self._data_broker

    def get_NAV(self) -> float:
        """Returns the net asset/liquidation value of the account."""
        return self.api.fetch_balance()[self.base_currency]["total"]

    def get_balance(self, instrument: str = None) -> float:
        """Returns account balance."""
        # TODO - check a pair hasnt been requested
        instrument = self.base_currency if instrument is None else instrument
        balances = self.api.fetch_balance()
        if instrument in balances:
            return balances[instrument]["total"]
        else:
            return 0

    def place_order(self, order: Order, **kwargs) -> None:
        """Place an order."""
        order()

        # Check order meets limits
        limits: dict = self.api.markets.get(order.instrument, {}).get("limits", {})
        if limits.get("amount") is not None:
            if float(order.size) < float(limits["amount"]["min"]):
                # Order too small
                self._logger.warning(f"Order below minimum size: {order}")
                return None

        # Add order params
        self._add_params(order)

        # Submit order to broker
        if order.order_type == "modify":
            placed_order = self._modify_order(order)

        else:
            # Regular order
            side = "buy" if order.direction > 0 else "sell"

            # Submit the order
            try:
                placed_order = self.api.create_order(
                    symbol=order.instrument,
                    type=order.order_type,
                    side=side,
                    amount=abs(order.size),
                    price=order.order_limit_price,
                    params=order.ccxt_params,
                )
            except Exception as e:
                placed_order = e

        return placed_order

    def get_orders(
        self, instrument: str = None, order_status: str = "open", **kwargs
    ) -> dict[str, Order]:
        """Returns orders associated with the account."""
        for _ in range(2):
            try:
                # Check for order id
                if "order_id" in kwargs:
                    # Fetch order by ID
                    if self.api.has["fetchOrder"]:
                        orders = [
                            self.api.fetch_order(
                                id=kwargs["order_id"], symbol=instrument
                            )
                        ]

                else:
                    # TODO - add exception handling
                    if order_status == "open":
                        # Fetch open orders (waiting to be filled)
                        orders = self.api.fetch_open_orders(instrument, **kwargs)

                    elif order_status == "cancelled":
                        # Fetch cancelled orders
                        orders = self.api.fetch_canceled_and_closed_orders(
                            instrument, **kwargs
                        )

                    elif order_status == "closed":
                        # Fetch closed orders
                        orders = self.api.fetch_closed_orders(instrument, **kwargs)

                    elif order_status == "conditional":
                        # Fetch conditional orders
                        orders = self.api.fetch_open_orders(
                            instrument, params={"orderType": "conditional"}
                        )

                    else:
                        # Unrecognised order status
                        raise Exception(f"Unrecognised order status '{order_status}'.")

                # Completed without exception, break loop
                break

            except ccxt.errors.NetworkError:
                # Throttle then try again
                time.sleep(1)

        # Convert
        orders = self._convert_list(orders, item_type="order")

        return orders

    def cancel_order(self, order_id: int, **kwargs) -> None:
        """Cancels order by order ID."""
        try:
            cancelled_order = self.api.cancel_order(id=order_id, **kwargs)

        except ccxt.errors.NetworkError:
            # Throttle then try again
            time.sleep(1)
            cancelled_order = self.api.cancel_order(id=order_id, **kwargs)

        except Exception as e:
            cancelled_order = e

        return cancelled_order

    def get_trades(self, instrument: str = None, **kwargs) -> dict:
        """Returns the open trades held by the account."""
        try:
            trades_list = self.api.fetch_my_trades(instrument, **kwargs)
        except ccxt.errors.NetworkError:
            # Throttle then try again
            time.sleep(1)
            trades_list = self.api.fetch_my_trades(instrument, **kwargs)

        # Convert to native Trades
        trades = self._convert_list(trades_list, item_type="trade")
        return trades

    def get_positions(self, instrument: str = None, **kwargs) -> dict[str, Position]:
        """Gets the current positions open on the account.

        Note that not all exchanges exhibit the same behaviour, and
        so caution must be taken when interpreting results. It is recommended
        to use the api directly and test with the exchange you plan to use
        to valid functionality.

        Parameters
        ----------
        instrument : str, optional
            The trading instrument name (symbol). The default is None.

        Returns
        -------
        open_positions : dict
            A dictionary containing details of the open positions.

        """
        for attempt in range(2):
            try:
                if instrument is None:
                    # Get all positions
                    if self.api.has["fetchPositions"]:
                        positions = self.api.fetch_positions(
                            symbols=None, params=kwargs
                        )
                        positions = self._convert_list(positions, item_type="position")

                    else:
                        raise Exception(
                            f"Exchange {self.exchange} does not have "
                            + "fetchPositions method."
                        )
                else:
                    # Get position in instrument provided
                    if self.api.has["fetchPosition"]:
                        position = self.api.fetch_position(instrument, params=kwargs)
                        if position is not None:
                            positions = {instrument: self._native_position(position)}
                        else:
                            positions = {}

                    elif self.api.has["fetchPositions"]:
                        positions = self.api.fetch_positions(
                            symbols=[instrument], params=kwargs
                        )
                        positions = self._convert_list(positions, item_type="position")
                    else:
                        raise Exception(
                            f"Exchange {self.exchange} does not have "
                            + "fetchPosition method."
                        )

                # Completed without exception, break loop
                break

            except ccxt.errors.NetworkError:
                # Throttle then try again
                time.sleep(1)

        # Check for zero-positions
        positions_dict = {}
        for symbol, pos in positions.items():
            if pos.net_position != 0:
                positions_dict[symbol] = pos

        return positions_dict

    def get_candles(
        self,
        instrument: str,
        granularity: str,
        count: int = None,
        start_time: datetime = None,
        end_time: datetime = None,
        **kwargs,
    ) -> pd.DataFrame:
        """Retrieves historical price data of a instrument from an exchange
        instance of the CCXT package.

        Parameters
        ----------
        instrument : str
            The instrument to fetch data for.

        granularity : str
            The candlestick granularity (eg. "1m", "15m", "1h", "1d").

        count : int, optional
            The number of candles to fetch (maximum 5000). The default is None.

        start_time : datetime, optional
            The data start time. The default is None.

        end_time : datetime, optional
            The data end time. The default is None.

        Returns
        -------
        data : DataFrame
            The price data, as an OHLCV DataFrame.
        """
        # Check requested start and end times
        if end_time is not None and end_time > datetime.now(tz=end_time.tzinfo):
            raise Exception("End time cannot be later than the current time.")

        if start_time is not None and start_time > datetime.now(tz=start_time.tzinfo):
            raise Exception("Start time cannot be later than the current time.")

        if start_time is not None and end_time is not None and start_time > end_time:
            raise Exception("Start time cannot be later than the end time.")

        # Check granularity was provided
        if granularity is None:
            granularity = "1m"

        def fetch_between_dates():
            # Fetches data between two dates
            max_count = 1000
            start_ts = int(start_time.timestamp() * 1000)
            end_ts = int(end_time.timestamp() * 1000)

            data = []
            while start_ts < end_ts:
                count = min(
                    max_count,
                    1
                    + (end_ts - start_ts)
                    / pd.Timedelta(granularity).total_seconds()
                    / 1000,
                )
                raw_data = self.api.fetch_ohlcv(
                    instrument,
                    timeframe=granularity,
                    since=start_ts,
                    limit=int(count),
                    params=kwargs,
                )
                # Append data
                data += raw_data

                # Increment start_ts
                start_ts = raw_data[-1][0] if len(raw_data) > 1 else end_ts

                # Sleep to throttle
                time.sleep(0.5)

            return data

        if count is not None:
            if start_time is None and end_time is None:
                # Fetch N most recent candles
                raw_data = self.api.fetch_ohlcv(
                    instrument, timeframe=granularity, limit=count, params=kwargs
                )
            elif start_time is not None and end_time is None:
                # Fetch N candles since start_time
                start_ts = (
                    None if start_time is None else int(start_time.timestamp() * 1000)
                )
                raw_data = self.api.fetch_ohlcv(
                    instrument,
                    timeframe=granularity,
                    since=start_ts,
                    limit=count,
                    params=kwargs,
                )
            elif end_time is not None and start_time is None:
                raise Exception(
                    "Fetching data from end_time and count is " + "not yet supported."
                )
            else:
                raw_data = fetch_between_dates()

        else:
            # Count is None
            try:
                assert start_time is not None and end_time is not None
                raw_data = fetch_between_dates()

            except AssertionError:
                raise Exception(
                    "When no count is provided, both start_time "
                    + "and end_time must be provided."
                )

        # Process data
        data = pd.DataFrame(
            raw_data, columns=["time", "Open", "High", "Low", "Close", "Volume"]
        ).set_index("time")
        data.index = pd.to_datetime(data.index, unit="ms")

        # TODO - normalise to UTC?

        return data

    def get_orderbook(
        self, instrument: str, limit: int = None, params: dict = None, **kwargs
    ) -> OrderBook:
        """Returns the orderbook"""
        params = params if params else {}
        response = self.api.fetch_order_book(
            symbol=instrument, limit=limit, params=params
        )

        # Unify format
        orderbook: dict[str, list] = {}
        for side in ["bids", "asks"]:
            orderbook[side] = []
            for level in response[side]:
                orderbook[side].append({"price": level[0], "size": level[1]})

        return OrderBook(instrument, orderbook)

    def get_public_trades(
        self,
        instrument: str,
        since: int = None,
        limit: int = None,
        params: dict = None,
        **kwargs,
    ):
        """Get the public trade history for an instrument."""
        params = params if params else {}
        ccxt_trades = self.api.fetch_trades(
            instrument, since=since, limit=limit, params=params
        )

        # Convert to standard form
        trades = []
        for trade in ccxt_trades:
            unified_trade = {
                "direction": 1 if trade["side"] == "buy" else -1,
                "price": float(trade["price"]),
                "size": float(trade["amount"]),
                "time": datetime.fromtimestamp(trade["timestamp"] / 1000),
            }
            trades.append(unified_trade)

        return trades

    def get_funding_rate(self, instrument: str, params: dict = None, **kwargs):
        """Returns the current funding rate."""
        params = params if params else {}
        response = self.api.fetch_funding_rate(instrument, params=params)
        fr_dict = {
            "symbol": instrument,
            "rate": response["fundingRate"],
            "time": response["fundingDatetime"],
        }
        return fr_dict

    def _ccxt_funding_history(
        self,
        instrument: str,
        count: int = None,
        start_time: datetime = None,
        end_time: datetime = None,
        params: dict = None,
    ):
        """Fetches the funding rate history."""
        params = params if params else {}

        def response_to_df(response):
            """Converts response to DataFrame."""
            times = []
            rates = []
            for chunk in response:
                times.append(pd.Timestamp(chunk["timestamp"], unit="ms"))
                rates.append(chunk["fundingRate"])
            return pd.DataFrame(data={"rate": rates}, index=times)

        def fetch_between_dates():
            # Fetches data between two dates
            count = 500
            start_ts = int(start_time.timestamp() * 1000)
            end_ts = int(end_time.timestamp() * 1000)

            rate_hist = pd.DataFrame()
            while start_ts <= end_ts:
                response = self.api.fetch_funding_rate_history(
                    symbol=instrument, since=start_ts, limit=count, params=params
                )

                # Append results
                df = response_to_df(response)
                rate_hist = pd.concat([rate_hist, df])

                # Increment start_ts
                start_ts = int(df.index[-1].timestamp() * 1000)

                # Sleep for API limit
                time.sleep(1)

            return rate_hist

        if count is not None:
            if start_time is None and end_time is None:
                # Fetch N most recent candles
                response = self.api.fetch_funding_rate_history(
                    symbol=instrument, limit=count, params=params
                )
                rate_hist = response_to_df(response)
            elif start_time is not None and end_time is None:
                # Fetch N candles since start_time
                start_ts = (
                    None if start_time is None else int(start_time.timestamp() * 1000)
                )
                response = self.api.fetch_funding_rate_history(
                    symbol=instrument, since=start_ts, limit=count, params=params
                )
                rate_hist = response_to_df(response)
            elif end_time is not None and start_time is None:
                raise Exception(
                    "Fetching data from end_time and count is " + "not yet supported."
                )
            else:
                rate_hist = fetch_between_dates()

        else:
            # Count is None
            if start_time is not None and end_time is not None:
                rate_hist = fetch_between_dates()
            else:
                response = self.api.fetch_funding_rate_history(
                    symbol=instrument,
                    params=params,
                )
                rate_hist = response_to_df(response)

        return rate_hist

    def get_trade_details(self, trade_ID: str) -> dict:
        """Returns the details of the trade specified by trade_ID."""
        raise NotImplementedError(
            "This method is not available, and will "
            + "be deprecated with a future release. Please use the "
            + "get_trades method instead."
        )

    def _add_params(self, order: Order):
        """Translates an order to add CCXT parameters for the specific exchange."""
        if self.exchange.lower() == "bybit":
            # https://bybit-exchange.github.io/docs/v5/order/create-order
            if order.take_profit:
                self._safe_add(order.ccxt_params, "takeProfit", order.take_profit)
                self._safe_add(order.ccxt_params, "tpslMode", "Partial")
            if order.stop_loss:
                self._safe_add(order.ccxt_params, "stopLoss", order.stop_loss)
                self._safe_add(order.ccxt_params, "tpslMode", "Partial")
        # TODO - support more exchanges

    @staticmethod
    def _safe_add(map: dict, key: str, value: any):
        """Adds a value to a map only if it is not already in there."""
        if key not in map:
            map[key] = value

    def _native_order(self, order: dict):
        """Returns a CCXT order as a native AutoTrader Order."""
        direction = 1 if order["side"] == "buy" else -1
        order_type = order["type"].lower()

        if order_type == "limit":
            limit_price = order["price"]
        else:
            limit_price = None

        stop_price = (
            float(order["stopPrice"]) if order["stopPrice"] is not None else None
        )

        native_order = Order(
            instrument=order["symbol"],
            direction=direction,
            order_type=order_type,
            status=order["status"],
            size=abs(order["amount"]),
            id=order["id"],
            order_limit_price=limit_price,
            order_stop_price=stop_price,
            order_time=datetime.fromtimestamp(order["timestamp"] / 1000),
            ccxt_order=order,
        )
        return native_order

    def _native_trade(self, trade):
        """Returns a CCXT trade as a native AutoTrader Trade."""
        direction = 1 if trade["side"] == "buy" else -1
        order_id_keys = ["orderID", "orderId2"]
        oid_assigned = False
        for key in order_id_keys:
            if key in trade["info"]:
                parent_order_id = trade["info"][key]
                oid_assigned = True
                break

        if not oid_assigned:
            parent_order_id = None

        native_trade = Trade(
            instrument=trade["symbol"],
            order_price=None,
            order_time=None,
            order_type=None,
            size=abs(trade["amount"]),
            last_price=None,
            fill_time=datetime.fromtimestamp(trade["timestamp"] / 1000).astimezone(
                timezone.utc
            ),
            fill_price=float(trade["price"]),
            fill_direction=direction,
            fee=trade["fee"]["cost"],
            id=trade["id"],
            order_id=parent_order_id,
        )

        return native_trade

    def _native_position(self, position):
        """Returns a CCXT position structure as a native
        AutoTrader Position.
        """
        # Get symbol
        try:
            symbol = position["symbol"]
        except:
            symbol = position["info"]["symbol"]

        direction = 1 if position["side"] == "long" else -1

        # Construct position object
        # TODO - add more attributes
        native_position = Position(
            instrument=symbol,
            net_position=position["contracts"] * direction,
            net_exposure=position["notional"],
            notional=position["notional"],
            pnl=position["unrealizedPnl"],
            PL=position["unrealizedPnl"],
            entry_price=position["entryPrice"],
            direction=direction,
            ccxt=position,
            avg_price=position["entryPrice"],
            total_margin=position["initialMargin"],
        )
        return native_position

    def _convert_list(self, items, item_type="order"):
        """Converts a list of trades or orders to a dictionary."""
        native_func = f"_native_{item_type}"
        id_key = "instrument" if item_type == "position" else "id"
        converted = {}
        for item in items:
            native = getattr(self, native_func)(item)
            converted[getattr(native, id_key)] = native
        return converted

    def _modify_order(self, order: Order):
        """Modify the size, type and price of an existing order."""
        side = "buy" if order.direction > 0 else "sell"
        try:
            modified_order = self.api.edit_order(
                id=order.related_orders[0],
                symbol=order.instrument,
                side=side,
                type=None,
                amount=order.size,
                price=order.order_limit_price,
                params=order.ccxt_params,
            )
        except Exception as e:
            modified_order = e
        return modified_order

    def get_precision(self, instrument: str, *args, **kwargs):
        """Returns the precision of the instrument."""
        # TODO - this method has been disabled as it was not properly handling precisions
        return None

        if instrument in self._instrument_precisions:
            # Precision already fetched, use stored value
            unified_response = self._instrument_precisions[instrument]
        else:
            # Fetch precision
            market = self._get_market(instrument)
            precision = market["precision"]

            size_precision = precision["amount"]
            price_precision = precision["price"]

            # Check for any decimals
            if "." in str(size_precision):
                size_precision = str(size_precision)[::-1].find(".")
            if "." in str(price_precision):
                price_precision = str(price_precision)[::-1].find(".")

            unified_response = {
                "size": size_precision,
                "price": price_precision,
            }

            # Store for later use
            self._instrument_precisions[instrument] = unified_response

        return unified_response

    def _get_market(self, instrument: str, *args, **kwargs):
        """Returns the raw get_market response from a CCXT exchange"""
        if instrument in self.markets:
            market = self.markets[instrument]
        elif instrument.split(":")[0] in self.markets:
            market = self.markets[instrument.split(":")[0]]
        elif f"{instrument.split('USDT')[0]}/USDT" in self.markets:
            market = self.markets[f"{instrument.split('USDT')[0]}/USDT"]
        else:
            raise Exception(
                f"{instrument} does not appear to be listed. "
                + "Please double check the naming."
            )
        return market

    def get_stepsize(self, instrument: str, *args, **kwargs):
        """Returns the stepsize for an instrument."""
        market = self._get_market(instrument)
        stepsize = float(market["limits"]["amount"]["min"])
        return stepsize

    def get_min_notional(self, instrument: str, *args, **kwargs):
        """Returns the minimum notional value a trade should hold."""
        market = self._get_market(instrument)
        min_notional = float(market["limits"]["cost"]["min"])
        return min_notional

    def get_ticksize(self, instrument: str, *args, **kwargs):
        """Returns the ticksize for an instrument."""
        market = self._get_market(instrument)
        try:
            ticksize = float(market["info"]["filters"][0]["tickSize"])
        except:
            raise Exception("Cannot retrieve ticksize.")
        return ticksize



from .autoplot import AutoPlot
from .autotrader import AutoTrader
from .utilities import TradeAnalysis, DataStream
from .brokers.trading import (
    Order,
    IsolatedPosition,
    Position,
    LimitOrder,
    MarketOrder,
    StopLimitOrder,
    Trade,
)

# Broker imports
# CCXT
try:
    from .brokers.ccxt import Broker as CCXT
except:
    pass

# Interactive Brokers
try:
    from .brokers.ib import Broker as IB
except:
    pass

# Oanda
try:
    from .brokers.oanda import Broker as Oanda
except:
    pass

# Virtual broker
from .brokers.virtual import Broker as VirtualBroker

# Define version number
__version__ = "1.1.1"



from __future__ import annotations
import pandas as pd
from autotrader.brokers.broker import Broker
from datetime import datetime, timezone, timedelta

try:
    import yfinance
except ImportError:
    raise Exception("Please install yfinance to use the Yahoo Finance data feed.")


class Broker(Broker):
    """Yahoo finance wrapper for data only."""

    def __init__(self, config: dict) -> None:
        self.api = yfinance.download

        # Create datastream object
        self._data_broker = self

    def __repr__(self):
        return "Yahoo Finance Broker Wrapper"

    def __str__(self):
        return super().__repr__()

    @property
    def data_broker(self):
        return self._data_broker

    def get_candles(
        self,
        instrument: str,
        granularity: str = None,
        count: int = None,
        start_time: datetime = None,
        end_time: datetime = None,
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        """Retrieves historical price data from yahoo finance.

        Parameters
        ----------
        instrument : str
            Ticker to dowload data for.

        granularity : str, optional
            The candlestick granularity. The default is None.

        count : int, optional
            The number of bars to fetch. The default is None.

        start_time : str, optional
            The start time as YYYY-MM-DD string or datetime object. The default
            is None.

        end_time : str, optional
            The end_time as YYYY-MM-DD string or datetime object. The default
            is None.

        Returns
        -------
        data : pd.DataFrame
            The price data, as an OHLC DataFrame.

        Notes
        -----
        If you are encountering a JSON error when using the yahoo finance API,
        try updating by running: pip install yfinance --upgrade --no-cache-dir

        Intraday data cannot exceed 60 days.
        """
        gran_map = {
            60: "1m",
            120: "2m",
            300: "5m",
            900: "15m",
            1800: "30m",
            3600: "60m",
            5400: "90m",
            3600: "1h",
            86400: "1d",
            432000: "5d",
            604800: "1wk",
            2419200: "1mo",
            7257600: "3mo",
        }
        try:
            granularity = gran_map[pd.Timedelta(granularity).total_seconds()]
        except KeyError:
            raise Exception(
                f"The specified granularity of '{granularity}' is not "
                + "valid for Yahoo Finance."
            )

        if count is not None and start_time is None and end_time is None:
            # Convert count to start and end dates (assumes end=now)
            end_time = datetime.now()
            start_time = end_time - timedelta(
                seconds=self._granularity_to_seconds(granularity, "yahoo") * 1.5 * count
            )

        # Fetch data
        data = self.api(
            tickers=instrument, start=start_time, end=end_time, interval=granularity
        )

        # Remove excess data
        if count is not None and start_time is None and end_time is None:
            data = data.tail(count)

        if data.index.tzinfo is None:
            # Data is naive, add UTC timezone
            data.index = data.index.tz_localize(timezone.utc)
        else:
            # Convert to UTC
            data.index = data.index.tz_convert(timezone.utc)

        return data

    def get_orderbook(self, instrument: str, *args, **kwargs):
        raise Exception("Orderbook data is not available from Yahoo Finance.")

    def get_public_trades(self, instrument: str, *args, **kwargs):
        raise Exception("Public trade data is not available from Yahoo Finance.")

    def _yahoo_quote_data(
        self,
        data: pd.DataFrame,
        pair: str,
        interval: str,
        from_date: datetime,
        to_date: datetime,
        count: int = None,
    ):
        """Returns nominal price data - quote conversion not supported for
        Yahoo finance API.
        """
        return data

    @staticmethod
    def _granularity_to_seconds(granularity: str, feed: str):
        """Converts the granularity to time in seconds."""
        # Note: will not work for week or month granularities
        letter = granularity[-1]
        number = float(granularity[:-1])
        conversions = {"m": 60, "h": 60 * 60, "d": 60 * 60 * 24}
        my_int = conversions[letter] * number
        return my_int



import numpy as np
import pandas as pd
from decimal import Decimal
from datetime import datetime
from typing import Literal, Union


class Order:
    """AutoTrader Order object."""

    def __init__(
        self,
        instrument: str = None,
        direction: int = None,
        size: float = None,
        order_type: Literal["market", "limit", "stop-limit", "modify"] = "market",
        order_limit_price: float = None,
        order_stop_price: float = None,
        stop_loss: float = None,
        stop_type: str = "limit",
        take_profit: float = None,
        related_orders: list[Union[str, int]] = None,
        **kwargs,
    ):
        """Create a new order.

        Parameters
        ----------
        instrument : str
            The trading instrument of the order.

        direction : int
            The direction of the order (1 for long, -1 for short).

        order_type : str
            The type of order. The default is 'market'.

        size : float
            The number of units.

        order_limit_price : float
            The limit price of the order (for 'limit' and 'stop-limit' order
            types).

        order_stop_price : float
            The stop price of the order (for 'stop-limit' order types).

        stop_loss : float
            The price to set the stop-loss at.

        stop_type : str
            The type of stop-loss (limit or trailing). The default is 'limit'.

        take_profit : float
            The price to set the take-profit at.

        related_orders : list[str | int]
            A list of related order/trade ID's.

        secType : str
            The security type (IB only).

        contract_month : str
            The contract month string (IB only).

        localSymbol : str
            The exchange-specific instrument symbol (IB only).

        limit_fee : str, optional
            The maximum fee to accept as a percentage (dYdX only). The default
            is '0.015'.

        ccxt_params : dict, optional
            The CCXT parameters dictionary to pass when creating an order. The
            default is {}.
        """
        # TODO - implement post-only

        # Required attributes
        self.instrument = instrument if instrument is not None else None
        self.direction = np.sign(direction) if direction is not None else None
        self.order_type = order_type

        # Optional arguments
        self.size = Decimal(str(size)) if size is not None else None
        self.order_price = None
        self.order_time = None
        self.order_limit_price = (
            Decimal(str(order_limit_price)) if order_limit_price else None
        )
        self.order_stop_price = (
            Decimal(str(order_stop_price)) if order_stop_price else None
        )

        # Multi-exchange handling
        self.exchange = None

        # Stop loss arguments
        self.stop_loss = Decimal(str(stop_loss)) if stop_loss else None
        self.stop_type = stop_type

        # Take profit arguments
        self.take_profit = Decimal(str(take_profit)) if take_profit else None

        self.related_orders = related_orders

        # Reduce only order
        self.parent_order = None  # Parent order ID
        self.reduce_only = False
        self.OCO = []  # One-cancels-other

        # IB attributes
        self.currency = None
        self.secType = None
        self.contract_month = None
        self.localSymbol = None

        # Oanda attributes
        self.trigger_price = "DEFAULT"

        # CCXT attributes
        self.ccxt_params = {}  # CCXT order parameters
        self.ccxt_order = {}  # CCXT native order structure

        # Meta
        self.reason = None
        self.strategy = None
        self.granularity = None
        self._sizing = None
        self._risk_pc = None
        self.id: Union[int, str] = None
        self.status: Literal["submitted", "pending", "open", "cancelled", "filled"] = (
            None
        )

        # Unpack kwargs
        for item in kwargs:
            setattr(self, item, kwargs[item])

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        if self.instrument is None:
            # Blank order
            return "Blank order"

        else:
            # Order constructed with instrument
            if self.size is not None:
                if self.direction is None:
                    return "Invalid order (direction not specified)"

                side = "buy" if self.direction > 0 else "sell"

                string = (
                    f"{self.size} "
                    + f"unit {self.instrument} {self.order_type} "
                    + f"{side} order"
                )

                # Append additional information
                if self.order_type == "limit":
                    if self.order_limit_price is None:
                        return "Invalid order (limit price not provided)"
                    string += f" @ {self.order_limit_price}"

                elif self.order_type == "stop-limit":
                    if self.order_limit_price is None:
                        return "Invalid order (limit price not provided)"
                    elif self.order_stop_price is None:
                        return "Invalid order (stop price not provided)"
                    string += f" @ {self.order_stop_price} / {self.order_limit_price}"

                elif self.order_type == "stop":
                    if self.order_stop_price is None:
                        return "Invalid order (stop price not provided)"
                    string += f" @ {self.order_stop_price}"

                return string

            else:
                # Size un-assigned
                return f"{self.instrument} {self.order_type} Order"

    def __call__(
        self,
        order_price: Decimal = None,
        order_time: datetime = datetime.now(),
    ) -> None:
        """Order object, called before submission to broker in
        autobot._qualify_orders.

        Parameters
        ----------
        order_price : Decimal, optional
            The order price. The default is None.

        order_time : datetime, optional
            The time of the order. The default is datetime.now().

        precision : dict, optional
            A dictionary containing the precision for order size and price.
            The default is None.

        Returns
        -------
        None
            Calling an Order will ensure all information is present.
        """
        self.order_price = (
            Decimal(str(order_price)) if order_price else self.order_price
        )
        self.order_time = order_time if order_time else self.order_time

        # Enforce size scalar
        self.size = abs(Decimal(str(self.size))) if self.size is not None else self.size
        self.status = "submitted"
        self.submitted = True

    @classmethod
    def _partial_fill(cls, order: "Order", units_filled: Decimal) -> "Order":
        """Partially fills the order."""
        # Enforce Decimal type
        units_filled = Decimal(str(units_filled))

        # Instantiate new order
        order_to_be_filled = cls()

        # Inherit attributes from base order
        for attribute, value in order.__dict__.items():
            setattr(order_to_be_filled, attribute, value)

        # Reset ID
        order_to_be_filled.id = None

        # Transfer units
        order_to_be_filled.size = units_filled
        order.size -= units_filled

        return order_to_be_filled

    def _check_precision(
        self,
    ):
        # TODO - implement
        raise NotImplementedError("This method has not been implemented yet.")

    def _validate(
        self,
    ):
        # TODO - add order validation method, ie. for IB, check all attributes are
        # assigned (eg. sectype, etc)
        raise NotImplementedError("This method has not been implemented yet.")

    def as_dict(self) -> dict:
        """Converts Order object to dictionary.

        Returns
        -------
        dict
            The order instance returned as a dict object.

        Notes
        -----
        This method enables legacy code operation, returning order/trade
        objects as a dictionary.
        """
        return self.__dict__

    @classmethod
    def _from_dict(cls, order_dict: dict) -> "Order":
        return Order(**order_dict)

    def _modify_from(self, order: "Order"):
        """Modify this order from another order."""
        # TODO - review what can be changed here

        # Check order limit price
        if order.order_limit_price is not None:
            self.order_limit_price = Decimal(str(order.order_limit_price))

        # Check size
        if order.size is not None:
            self.size = Decimal(str(order.size))

        # Check stop loss
        if order.stop_loss is not None:
            self.stop_loss = Decimal(str(order.stop_loss))

        # Check take profit
        if order.take_profit is not None:
            self.take_profit = Decimal(str(order.take_profit))


class MarketOrder(Order):
    """Market order type."""

    def __init__(
        self,
        instrument: str = None,
        direction: int = None,
        size: float = None,
        **kwargs,
    ):
        # Create base Order
        super().__init__(
            instrument=instrument,
            direction=direction,
            order_type="market",
            size=size,
            **kwargs,
        )


class LimitOrder(Order):
    """Limit order type."""

    def __init__(
        self,
        instrument: str = None,
        direction: int = None,
        size: float = None,
        order_limit_price: float = None,
        **kwargs,
    ):
        # Create base Order
        super().__init__(
            instrument=instrument,
            direction=direction,
            order_type="limit",
            size=size,
            order_limit_price=order_limit_price,
            **kwargs,
        )


class StopLimitOrder(Order):
    """Stop-limit order type."""

    def __init__(
        self,
        instrument: str = None,
        direction: int = None,
        size: float = None,
        order_limit_price: float = None,
        order_stop_price: float = None,
        **kwargs,
    ):
        # Create base Order
        super().__init__(
            instrument=instrument,
            direction=direction,
            order_type="stop-limit",
            size=size,
            order_limit_price=order_limit_price,
            order_stop_price=order_stop_price,
            **kwargs,
        )


class IsolatedPosition(Order):
    """AutoTrader IsolatedPosition. Use to connect SL and TP orders to individual
    trades.

    Attributes
    ----------
    unrealised_PL : float
        The floating PnL of the trade.

    margin_required : float
        The margin required to maintain the trade.

    time_filled : datetime
        The time at which the trade was filled.

    fill_price : float
        The price at which the trade was filled.

    last_price : float
        The last price observed for the instrument associated with the trade.

    last_time : datetime
        The last time observed for the instrument associated with the trade.

    exit_price : float
        The price at which the trade was closed.

    exit_time : datetime
        The time at which the trade was closed.

    fees : float
        The fees associated with the trade.

    parent_id : int
        The ID of the order which spawned the trade.

    id : int
        The trade ID.

    status : str
        The status of the trade (open or closed).

    split : bool
        If the trade has been split.

    Notes
    -----
    When a trade is created from an Order, the Order will be marked as filled.
    """

    def __init__(self, order: Order = None, **kwargs):
        # Trade data
        self.unrealised_PL = 0
        self.margin_required = 0
        self.time_filled = None
        self.fill_price = None

        self.last_price = None
        self.last_time = None

        self.profit = 0
        self.balance = None
        self.exit_price = None
        self.exit_time = None
        self.fees = None

        # Meta data
        self.parent_id = None  # ID of order which spawned trade
        self.id = None
        self.status = None  # options: open -> closed
        self.split = False

        # Unpack kwargs
        for item in kwargs:
            setattr(self, item, kwargs[item])

        # Inherit order attributes
        if order:
            self._inheret_order(order)
            self.parent_id = order.id

    def __repr__(self):
        direction = "long" if self.direction > 0 else "short"
        return (
            f"{round(self.size,3)} unit {direction} {self.instrument} IsolatedPosition"
        )

    def __str__(self):
        return "AutoTrader IsolatedPosition"

    def _inheret_order(self, order: Order) -> None:
        for attribute, value in order.__dict__.items():
            setattr(self, attribute, value)

    @classmethod
    def _split(
        cls, trade: "IsolatedPosition", split_units: Decimal
    ) -> "IsolatedPosition":
        """Splits parent IsolatedPosition into new object for partial
        closures.

        split units are given to the new trade.
        """
        split_trade = cls()
        for attribute, value in trade.__dict__.items():
            setattr(split_trade, attribute, value)

        # Reset ID
        split_trade.parent_id = trade.parent_id
        split_trade.order_id = None

        # Transfer units
        split_trade.size = Decimal(str(split_units))
        trade.size -= Decimal(str(split_units))

        # Mark original trade as split
        trade.split = True

        return split_trade


class Trade:
    """AutoTrader Trade object. Represents an exchange of value."""

    def __init__(
        self,
        instrument: str,
        order_price: Decimal,
        order_time: datetime,
        order_type: str,
        size: Decimal,
        last_price: Decimal,
        fill_time: datetime,
        fill_price: Decimal,
        fill_direction: int,
        fee: Decimal,
        **kwargs,
    ):
        """Trade constructor."""
        # Trade data
        self.fill_time = fill_time
        self.fill_price = fill_price
        self.direction = fill_direction
        self.fee = fee
        self.last_price = last_price
        self.order_price = order_price
        self.order_time = order_time
        self.order_type = order_type
        self.size = size
        self.instrument = instrument

        # Precision attributes
        # TODO - review this!
        self._price_precision = 15
        self._size_precision = 15

        # Meta-data
        self.id = None
        self.order_id = None

        for item in kwargs:
            setattr(self, item, kwargs[item])

    def __repr__(self):
        direction = "long" if self.direction > 0 else "short"
        return f"{round(self.size,3)} unit {direction} {self.instrument} trade @ {self.fill_price}"

    def __str__(self):
        return "AutoTrader Trade"


class Position:
    """AutoTrader Position object.

    Attributes
    ----------
    instrument : str
        The trade instrument of the position.

    pnl : float
        The pnl of the position.

    long_units : float
        The number of long units in the position.

    long_PL : float
        The PnL of the long units in the position.

    long_margin : float
        The margin required to maintain the long units in the position.

    short_units : float
        The number of short units in the position.

    short_PL : float
        The PnL of the short units in the position.

    short_margin : float
        The margin required to maintain the short units in the position.

    total_margin : float
        The total margin required to maintain the position.

    trade_IDs : list[int]
        The trade ID's associated with the position.

    net_position : float
        The total number of units in the position.

    net_exposure : float
        The net exposure (in $ value) of the position.

    PL : float
        The floating PnL (IB only).

    contracts : list
        The contracts associated with the position (IB only).

    portfolio_items : list
        The portfolio items associated with the position (IB only).
    """

    def __init__(self, **kwargs):
        self.instrument = None
        self.net_position = None
        self.pnl = 0
        self.long_units = None
        self.long_PL = None
        self.long_margin = None
        self.short_units = None
        self.short_PL = None
        self.short_margin = None
        self.total_margin = None
        self.trade_IDs = None
        self.net_exposure = None
        self.notional = 0
        # TODO - review this!
        self.price_precision = 15
        self.size_precision = 15
        self.avg_price = None
        self._prev_avg_price = None

        # TODO - include position open time, close time
        # Also need to track max units ? I think that can all be built
        # from fills
        self.entry_time = None
        self.last_price = None
        self.last_time = None
        self.exit_time = None
        self.exit_price = None

        self.direction = None
        # self.max_size = None # ?

        # IB Attributes
        self.PL = None
        self.contracts = None
        self.portfolio_items = None

        # dYdX Attributes
        self.entry_price = None

        # CCXT Attributes
        self.ccxt = None

        for item in kwargs:
            setattr(self, item, kwargs[item])

    def __repr__(self):
        return f"Position in {self.instrument}"

    def __str__(self):
        return "AutoTrader Position"

    def _update_with_fill(self, trade: Trade):
        """Updates the position with the order provided."""
        net_position_before_fill = self.net_position

        # Update net position
        self.net_position += trade.size * trade.direction

        # Update average entry price
        self._prev_avg_price = self.avg_price
        if self.net_position * net_position_before_fill >= 0:
            # Position has not flipped
            if abs(self.net_position) > abs(net_position_before_fill):
                # Position has increased
                self.avg_price = round(
                    (
                        self.avg_price * abs(self.net_position)
                        + trade.fill_price * trade.size
                    )
                    / (abs(self.net_position) + trade.size),
                    self.price_precision,
                )
            # If position has reduced, average entry price will not change
        else:
            # Position has flipped
            self.avg_price = round(trade.fill_price, self.price_precision)

        # Update last price and last time
        self.last_price = trade.last_price
        self.last_time = trade.fill_time

        # TODO - update value of position
        # self.net_exposure
        self.notional = self.last_price * abs(self.net_position)

    @classmethod
    def _from_fill(cls, trade: Trade):
        """Returns a Position from a fill."""
        position = cls(
            instrument=trade.instrument,
            net_position=abs(trade.size) * trade.direction,
            last_price=trade.last_price,
            last_time=trade.fill_time,
            entry_time=trade.fill_time,
            entry_price=trade.fill_price,
            avg_price=trade.fill_price,
            notional=trade.fill_price * abs(trade.size),
            price_precision=trade._price_precision,
            size_precision=trade._size_precision,
            direction=trade.direction,
        )
        # TODO - update attributes created
        # - exposure, value, etc.
        return position

    def as_dict(self) -> dict:
        """Converts Position object to dictionary.

        Returns
        -------
        dict
            The Position instance returned as a dict object.

        Notes
        -----
        This method enables legacy code operation, returning order/trade
        objects as a dictionary.
        """
        return self.__dict__


class OrderBook:
    def __init__(self, instrument: str, initial_state: dict):
        # TODO - review, make all Decimals
        self.instrument = instrument
        self.bids = None
        self.asks = None
        self._midprice = None
        self._spread = None

        # Initialise from initial state
        self.bids = pd.DataFrame(initial_state["bids"]).astype(float)
        self.asks = pd.DataFrame(initial_state["asks"]).astype(float)

        # Sort quotes
        self.bids.sort_values(by="price", ascending=False, inplace=True)
        self.asks.sort_values(by="price", ascending=True, inplace=True)

        # Calculate spread and midprice
        spread = float(self.asks.price.min()) - float(self.bids.price.max())
        midprice = (float(self.asks.price.min()) + float(self.bids.price.max())) / 2

        # Quantize
        # TODO - use ticksize and step size to quantize
        ref = Decimal(str(self.bids["price"][0]))
        self.spread = float(Decimal(spread).quantize(ref))
        self.midprice = float(Decimal(midprice).quantize(ref))

    def __repr__(self):
        return f"{self.instrument} Order Book snapshot"

    @property
    def midprice(self):
        return self._midprice

    @midprice.setter
    def midprice(self, value):
        self._midprice = value

    @property
    def spread(self):
        return self._spread

    @spread.setter
    def spread(self, value):
        self._spread = value



import os
import sys
import yaml
import time
import pickle
import logging
import autotrader
import numpy as np
import pandas as pd
from art import tprint
from abc import abstractmethod
from typing import Union, Optional
from datetime import datetime, timedelta
from prometheus_client import start_http_server, Gauge
from autotrader.brokers.broker import AbstractBroker, Broker

try:
    from ccxt_download.utilities import load_data
    from ccxt_download.constants import DEFAULT_DOWNLOAD_DIR, CANDLES
except:
    pass


def read_yaml(file_path: str) -> dict:
    """Function to read and extract contents from .yaml file.

    Parameters
    ----------
    file_path : str
        The absolute filepath to the yaml file.

    Returns
    -------
    dict
        The loaded yaml file in dictionary form.
    """
    with open(file_path, "r") as f:
        return yaml.safe_load(f)


def write_yaml(data: dict, filepath: str) -> None:
    """Writes a dictionary to a yaml file.

    Parameters
    ----------
    data : dict
        The dictionary to write to yaml.

    filepath : str
        The filepath to save the yaml file.

    Returns
    -------
    None
        The data will be written to the filepath provided.
    """
    with open(filepath, "w") as outfile:
        yaml.dump(data, outfile, default_flow_style=False)


def print_banner():
    tprint("AutoTrader", font="tarty1")


def get_broker_config(
    broker: str,
    global_config: Optional[dict] = None,
    environment: Optional[str] = "paper",
) -> dict:
    """Returns a broker configuration dictionary.

    Parameters
    ----------
    broker : str
        The name(s) of the broker/exchange. Specify multiple exchanges using
        comma separation.

    global_config : dict
        The global configuration dictionary.

    environment : str, optional
        The trading evironment ('demo' or 'real').

    """
    all_config = {}
    inputted_brokers = broker.lower().replace(" ", "").split(",")

    # Check global_config
    if global_config is None:
        try:
            global_config = read_yaml("config/keys.yaml")
        except:
            pass

    for broker in inputted_brokers:
        # Check for CCXT
        if broker.split(":")[0].lower() == "ccxt":
            broker_key = broker
            broker, exchange = broker.lower().split(":")
        elif broker.split(":")[0].lower() == "virtual":
            broker_key = ":".join(broker.split(":")[1:])
            broker = "virtual"
        else:
            broker_key = broker

        supported_brokers = ["oanda", "ib", "ccxt", "virtual"]
        if broker.lower() not in supported_brokers:
            raise Exception(f"Unsupported broker: '{broker}'")

        if broker != "ccxt" and environment.lower() not in ["live", "paper"]:
            raise Exception("Trading environment must either be 'live' or 'paper'.")

        # Live trading
        if broker.lower() == "oanda":
            api_key = "LIVE" if environment.lower() == "live" else "PRACTICE"
            oanda_conf = global_config["OANDA"]

            # Initialise config dict
            config = {"PORT": oanda_conf["PORT"]}

            # Unpack
            if f"{api_key}_API" in oanda_conf:
                config["API"] = oanda_conf[f"{api_key}_API"]
            else:
                raise Exception(
                    f"Please define {api_key}_API in your "
                    + f"account configuration for {environment} trading."
                )

            if f"{api_key}_ACCESS_TOKEN" in oanda_conf:
                config["ACCESS_TOKEN"] = oanda_conf[f"{api_key}_ACCESS_TOKEN"]
            else:
                raise Exception(
                    f"Please define {api_key}_ACCESS_TOKEN in "
                    + f"your account configuration for {environment} trading."
                )

            config["ACCOUNT_ID"] = (
                oanda_conf["DEFAULT_ACCOUNT_ID"]
                if "custom_account_id" not in global_config
                else global_config["custom_account_id"]
            )

        elif broker.lower() == "ib":
            config = {
                "host": (
                    global_config["host"] if "host" in global_config else "127.0.0.1"
                ),
                "port": global_config["port"] if "port" in global_config else 7497,
                "clientID": (
                    global_config["clientID"] if "clientID" in global_config else 1
                ),
                "account": (
                    global_config["account"] if "account" in global_config else ""
                ),
                "read_only": (
                    global_config["read_only"]
                    if "read_only" in global_config
                    else False
                ),
            }

        elif broker.lower() == "ccxt":
            if global_config is not None and broker_key.upper() in global_config:
                # Use configuration provided in config
                config_data = global_config[broker_key.upper()]

            else:
                # Use public-only connection
                config_data = {}

            # Select config based on environment
            if environment.lower() in config_data:
                config_data = config_data[environment.lower()]
            elif "mainnet" in config_data and environment.lower() == "live":
                config_data = config_data["mainnet"]
            elif "testnet" in config_data and environment.lower() == "paper":
                config_data = config_data["testnet"]

            api_key = config_data["api_key"] if "api_key" in config_data else None
            secret = config_data["secret"] if "secret" in config_data else None
            currency = (
                config_data["base_currency"]
                if "base_currency" in config_data
                else "USDT"
            )
            sandbox_mode = False if environment.lower() == "live" else True
            config = {
                "data_source": "ccxt",
                "exchange": exchange,
                "api_key": api_key,
                "secret": secret,
                "sandbox_mode": sandbox_mode,
                "base_currency": currency,
            }
            other_args = {"options": {}, "password": None}
            for key, default_val in other_args.items():
                if key in config_data:
                    config[key] = config_data[key]
                else:
                    config[key] = default_val

        elif broker.lower() == "virtual":
            config = {}

        else:
            raise Exception(f"No configuration available for {broker}.")

        # Append to full config
        all_config[broker_key] = config

    # Check length
    if len(all_config) == 1:
        all_config = config

    return all_config


def get_data_config(feed: str, global_config: Optional[dict] = None, **kwargs) -> dict:
    """Returns a data configuration dictionary for AutoData.
    Parameters
    ----------
    feed : str
        The name of the data feed.

    global_config : dict
        The global configuration dictionary.
    """
    # TODO - review if this is needed - probably can merge with broker config above
    if feed is None:
        print("Please specify a data feed.")
        sys.exit(0)

    # Check for CCXT
    if feed.split(":")[0].lower() == "ccxt":
        feed, exchange = feed.lower().split(":")

    # Check feed
    supported_feeds = ["oanda", "ib", "ccxt", "yahoo", "local", "none"]
    if feed.lower() not in supported_feeds:
        raise Exception(f"Unsupported data feed: '{feed}'")

    # Check global_config
    if global_config is None:
        try:
            global_config = read_yaml("config/keys.yaml")
        except:
            pass

    # Check for required authentication
    auth_feeds = ["oanda", "ib"]
    if feed.lower() in auth_feeds and global_config is None:
        raise Exception(
            f"Data feed '{feed}' requires authentication. "
            + "Please provide authentication details in the global config."
        )

    # Construct configuration dict
    config = {"data_source": feed.lower()}

    if feed.lower() == "oanda":
        environment = kwargs["environment"] if "environment" in kwargs else "paper"
        api_key = "LIVE" if environment.lower() == "live" else "PRACTICE"
        oanda_conf = global_config["OANDA"]

        # Unpack
        if f"{api_key}_API" in oanda_conf:
            config["API"] = oanda_conf[f"{api_key}_API"]
        else:
            raise Exception(
                f"Please define {api_key}_API in your "
                + f"account configuration for {environment} trading."
            )

        if f"{api_key}_ACCESS_TOKEN" in oanda_conf:
            config["ACCESS_TOKEN"] = oanda_conf[f"{api_key}_ACCESS_TOKEN"]
        else:
            raise Exception(
                f"Please define {api_key}_ACCESS_TOKEN in "
                + f"your account configuration for {environment} trading."
            )

        config["PORT"] = oanda_conf["PORT"]
        config["ACCOUNT_ID"] = (
            oanda_conf["DEFAULT_ACCOUNT_ID"]
            if "custom_account_id" not in global_config
            else global_config["custom_account_id"]
        )

    elif feed.lower() == "ib":
        config["host"] = (
            global_config["host"] if "host" in global_config else "127.0.0.1"
        )
        config["port"] = global_config["port"] if "port" in global_config else 7497
        config["clientID"] = (
            global_config["clientID"] if "clientID" in global_config else 1
        )
        config["account"] = (
            global_config["account"] if "account" in global_config else ""
        )
        config["read_only"] = (
            global_config["read_only"] if "read_only" in global_config else False
        )

    elif feed.lower() == "ccxt":
        # Try add authentication with global config
        if global_config is not None:
            # Global config is available
            try:
                # Try get api keys
                environment = (
                    kwargs["environment"] if "environment" in kwargs else "paper"
                )
                config = get_broker_config(
                    broker=f"{feed.lower()}:{exchange}",
                    global_config=global_config,
                    environment=environment,
                )
            except:
                # Didn't work, just set exchange
                config["exchange"] = exchange
        else:
            # No global config available, just set exchange
            config["exchange"] = exchange

    return config


def get_streaks(trade_summary):
    """Calculates longest winning and losing streaks from trade summary."""
    profit_list = trade_summary[trade_summary["status"] == "closed"].profit.values
    longest_winning_streak = 1
    longest_losing_streak = 1
    streak = 1

    for i in range(1, len(profit_list)):
        if np.sign(profit_list[i]) == np.sign(profit_list[i - 1]):
            streak += 1

            if np.sign(profit_list[i]) > 0:
                # update winning streak
                longest_winning_streak = max(longest_winning_streak, streak)
            else:
                # Update losing
                longest_losing_streak = max(longest_losing_streak, streak)

        else:
            streak = 1

    return longest_winning_streak, longest_losing_streak


def unpickle_broker(picklefile: Optional[str] = ".virtual_broker"):
    """Unpickles a virtual broker instance for post-processing."""
    with open(picklefile, "rb") as file:
        instance = pickle.load(file)
    return instance


class CustomLoggingFormatter(logging.Formatter):
    """Custom logging formatter (for StreamHandlers only)."""

    grey = "\x1b[38;21m"
    blue = "\x1b[38;5;39m"
    yellow = "\x1b[38;5;226m"
    red = "\x1b[38;5;196m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    default_fmt = "%(asctime)s | %(levelname)8s | %(name)8s | %(message)s (%(filename)s:%(lineno)d)"

    def __init__(self, fmt=None):
        super().__init__()

        self.fmt = fmt if fmt else self.default_fmt
        self.FORMATS = {
            logging.DEBUG: self.grey + self.fmt + self.reset,
            logging.INFO: self.blue + self.fmt + self.reset,
            logging.WARNING: self.yellow + self.fmt + self.reset,
            logging.ERROR: self.red + self.fmt + self.reset,
            logging.CRITICAL: self.bold_red + self.fmt + self.reset,
        }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def get_logger(
    name: str,
    stdout: Optional[bool] = True,
    stdout_level: Optional[Union[int, str]] = logging.INFO,
    file: Optional[bool] = False,
    file_level: Optional[Union[int, str]] = logging.INFO,
    log_dir: Optional[str] = "autotrader_logs",
):
    """Get (or create) a logger."""
    # Create logger
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    # Check for and clear any existing handlers on this logger
    if logger.hasHandlers():
        logger.handlers.clear()

    # Create new handlers
    handlers = []

    # Check for logging to file
    if file:
        # Check log directory exists
        if not os.path.exists(log_dir):
            os.mkdir(log_dir)
        logfile_path = os.path.join(log_dir, f"{name.replace(os.sep, '_')}.log")
        fh = logging.FileHandler(logfile_path, mode="a")
        fh.setLevel(file_level)
        fh.setFormatter(logging.Formatter(CustomLoggingFormatter.default_fmt))
        handlers.append(fh)

    # Check for logging to stdout
    if stdout:
        sh = logging.StreamHandler()
        stdout_level = stdout_level if stdout_level else logging.ERROR
        sh.setLevel(stdout_level)
        sh.setFormatter(CustomLoggingFormatter())
        handlers.append(sh)

    # Add handler to logger
    for h in handlers:
        logger.addHandler(h)

    return logger


class TradeAnalysis:
    """AutoTrader trade analysis class.

    Attributes
    ----------
    instruments_traded : list
        The instruments traded during the trading period.

    account_history : pd.DataFrame
        A timeseries history of the account during the trading period.

    holding_history : pd.DataFrame
        A timeseries summary of holdings during the trading period, by portfolio
        allocation fraction.

    order_history : pd.DataFrame
        A timeseries history of orders placed during the trading period.

    cancelled_orders : pd.DataFrame
        Orders which were cancelled during the trading period.

    trade_history : pd.DataFrame
        A history of all trades (fills) made during the trading period.

    """

    def __init__(
        self,
        broker,
        broker_histories: dict,
        instrument: str = None,
        price_history: pd.DataFrame = None,
    ):
        # Meta data
        self.brokers_used = None
        self.broker_results = None
        self.instruments_traded = None

        # Histories
        self.account_history = None
        self.position_history = None
        self.position_summary = None
        self.order_history = None
        self.cancelled_orders = None
        self.trade_history = None
        self.price_history = price_history

        # Perform analysis
        self.analyse_account(broker, broker_histories, instrument)

    def __str__(self):
        return "AutoTrader Trading Results"

    def __repr__(self):
        return "AutoTrader Trading Results"

    def analyse_account(
        self,
        broker,
        broker_histories: dict,
        instrument: Optional[str] = None,
    ) -> None:
        """Analyses trade account and creates summary of key details."""
        if not isinstance(broker, dict):
            # Single broker - create dummy dict
            broker_instances = {list(broker_histories.keys())[0]: broker}
        else:
            # Multiple brokers passed in as dict
            broker_instances = broker

        # Process results from each broker instance
        broker_results = {}
        for broker_name, broker in broker_instances.items():
            # Construct trade and order summaries
            all_orders = {}
            for status in ["pending", "open", "filled", "cancelled"]:
                orders = broker.get_orders(order_status=status)
                all_orders.update(orders)

            orders = TradeAnalysis.create_trade_summary(
                orders=all_orders, instrument=instrument, broker_name=broker_name
            )
            trade_history = TradeAnalysis.create_fill_summary(
                fills=broker._fills, broker_name=broker_name
            )

            account_history = pd.DataFrame(
                data=broker_histories[broker_name],
            )
            account_history.set_index("time", inplace=True)

            position_history = TradeAnalysis.create_position_history(
                trade_history=trade_history,
                account_history=account_history,
            )

            position_summary = TradeAnalysis._create_position_summary(
                broker._positions, broker._closed_positions
            )

            # Calculate drawdown
            account_history["drawdown"] = (
                account_history.NAV / account_history.NAV.cummax() - 1
            )

            # Save results for this broker instance
            broker_results[broker_name] = {
                "instruments_traded": list(orders.instrument.unique()),
                "account_history": self._decimal_to_float(account_history),
                "position_history": self._decimal_to_float(position_history),
                "position_summary": self._decimal_to_float(position_summary),
                "order_history": self._decimal_to_float(orders),
                "cancelled_orders": self._decimal_to_float(
                    orders[orders.status == "cancelled"]
                ),
                "trade_history": self._decimal_to_float(trade_history),
            }

        # Save all results
        self.broker_results = broker_results

        # Aggregate across broker instances
        self._aggregate_across_brokers(broker_results)

    @staticmethod
    def _decimal_to_float(df: pd.DataFrame):
        """Cast all numeric types to floats to support plotting."""
        for col in df:
            try:
                df[col] = df[col].astype(float)
            except:
                pass
        return df

    @staticmethod
    def create_position_history(
        trade_history: pd.DataFrame,
        account_history: pd.DataFrame,
    ) -> pd.DataFrame:
        """Creates a history of positions held, recording number of units held
        at each timestamp.
        """
        # Use fills to reconstruct position history, in terms of units held
        instruments_traded = list(trade_history["instrument"].unique())

        position_histories_dict = {}
        for instrument in instruments_traded:
            instrument_trade_hist = trade_history[
                trade_history["instrument"] == instrument
            ]
            directional_trades = (
                instrument_trade_hist["direction"] * instrument_trade_hist["size"]
            )
            net_position_hist = directional_trades.cumsum()

            # Filter out duplicates
            net_position_hist = net_position_hist[
                ~net_position_hist.index.duplicated(keep="last")
            ]

            # Reindex to account history index
            # TODO - match index timezone info
            net_position_hist = net_position_hist[~net_position_hist.index.isna()]
            try:
                net_position_hist = net_position_hist.reindex(
                    index=account_history.index, method="ffill"
                ).fillna(0)
            except:
                print("Could not reindex position history on account history index.")

            # Save result
            position_histories_dict[instrument] = net_position_hist

        # If no trades, create empty dataframe
        if len(instruments_traded) == 0:
            position_histories_dict = {
                "empty": pd.DataFrame(index=account_history.index)
            }

        position_histories = pd.concat(position_histories_dict, axis=1)
        return position_histories

    @staticmethod
    def _create_position_summary(open_positions, closed_positions):
        """Creates a summary of positions held."""
        # # Analyse closed positions (currently unused)
        # open_positions_summary = {}
        # for instrument, position in open_positions.items():
        #     directions = [p.direction for p in positions]

        #     # Save analysis
        #     open_positions_summary[instrument] = {
        #         "directions": directions,
        #         "no_long": directions.count(1),
        #         "no_short": directions.count(-1),
        #     }

        # Analyse closed positions
        closed_positions_results = {}
        closed_positions_summary = {}
        for instrument, positions in closed_positions.items():
            directions = [p.direction for p in positions]
            durations = [p.exit_time - p.entry_time for p in positions]

            # Save analysis
            closed_positions_results[instrument] = {
                "directions": directions,
                "durations": durations,
            }
            closed_positions_summary[instrument] = {
                "no_long": directions.count(1),
                "no_short": directions.count(-1),
            }
            if len(durations) > 0:
                closed_positions_summary[instrument]["avg_duration"] = np.mean(
                    durations
                )
                if directions.count(1) > 0:
                    closed_positions_summary[instrument]["avg_long_duration"] = np.mean(
                        np.array(durations)[np.array(directions) == 1]
                    )
                else:
                    closed_positions_summary[instrument]["avg_long_duration"] = None
                if directions.count(-1) > 0:
                    closed_positions_summary[instrument]["avg_short_duration"] = (
                        np.mean(np.array(durations)[np.array(directions) == -1])
                    )
                else:
                    closed_positions_summary[instrument]["avg_short_duration"] = None

        # Create summary dataframe
        summary = pd.DataFrame(
            data=closed_positions_summary,
        )

        return summary

    def _aggregate_across_brokers(self, broker_results):
        """Aggregates trading history across all broker instances."""
        brokers_used = []
        instruments_traded = []
        cancelled_orders = pd.DataFrame()
        position_history = pd.DataFrame()
        position_summary = pd.DataFrame()
        order_history = pd.DataFrame()
        trade_history = pd.DataFrame()
        account_history = None
        for broker, results in broker_results.items():
            orders = results["order_history"]
            brokers_used.append(broker)

            # Append unique instruments traded
            unique_instruments = orders.instrument.unique()
            [
                (
                    instruments_traded.append(instrument)
                    if instrument not in instruments_traded
                    else None
                )
                for instrument in unique_instruments
            ]

            # Aggregate account history
            if account_history is None:
                # Initialise
                account_history = results["account_history"]
            else:
                # Reindex each dataset
                original_index = results["account_history"].reindex(
                    index=account_history.index, method="ffill"
                )
                new_index = account_history.reindex(
                    index=results["account_history"].index, method="ffill"
                )
                if len(original_index) >= len(new_index):
                    # Use original index
                    account_history += original_index
                else:
                    # Use new index
                    account_history = new_index + results["account_history"]

            # Concatenate trades, orders and trade_history
            cancelled_orders = pd.concat(
                [cancelled_orders, results["cancelled_orders"]]
            )
            order_history = pd.concat([order_history, results["order_history"]])
            trade_history = pd.concat([trade_history, results["trade_history"]])
            position_history = pd.concat(
                [position_history, results["position_history"]]
            )
            # TODO - implement position summary for multiple exchanges
            position_summary = results["position_summary"]

        # Assign attributes
        self.brokers_used = brokers_used
        self.instruments_traded = instruments_traded
        self.account_history = account_history
        self.position_history = position_history
        self.position_summary = position_summary
        self.order_history = order_history
        self.cancelled_orders = cancelled_orders
        self.trade_history = trade_history

    @staticmethod
    def create_fill_summary(fills: list, broker_name: str = None):
        """Creates a dataframe of fill history."""
        # Initialise lists
        fill_dict = {
            "order_time": [],
            "order_price": [],
            "order_type": [],
            "fill_time": [],
            "fill_price": [],
            "direction": [],
            "size": [],
            "fee": [],
            "instrument": [],
            "id": [],
            "order_id": [],
        }
        for fill in fills:
            fill_dict["order_time"].append(fill.order_time)
            fill_dict["order_price"].append(fill.order_price)
            fill_dict["fill_time"].append(fill.fill_time)
            fill_dict["fill_price"].append(fill.fill_price)
            fill_dict["direction"].append(fill.direction)
            fill_dict["size"].append(fill.size)
            fill_dict["fee"].append(fill.fee)
            fill_dict["instrument"].append(fill.instrument)
            fill_dict["id"].append(fill.id)
            fill_dict["order_id"].append(fill.order_id)
            fill_dict["order_type"].append(fill.order_type)

        fill_df = pd.DataFrame(data=fill_dict, index=fill_dict["fill_time"])
        fill_df["broker"] = broker_name

        return fill_df

    @staticmethod
    def create_trade_summary(
        trades: dict = None,
        orders: dict = None,
        instrument: str = None,
        broker_name: str = None,
    ) -> pd.DataFrame:
        """Creates a summary dataframe for trades and orders."""
        # TODO - review this, could likely be cleaner
        instrument = None if isinstance(instrument, list) else instrument

        if trades is not None:
            iter_dict = trades
        else:
            iter_dict = orders

        iter_dict = {} if iter_dict is None else iter_dict

        product = []
        status = []
        ids = []
        times_list = []
        order_type = []
        order_price = []
        size = []
        direction = []
        stop_price = []
        take_price = []
        reason = []

        if trades is not None:
            entry_time = []
            fill_price = []
            profit = []
            portfolio_balance = []
            exit_times = []
            exit_prices = []
            trade_duration = []
            fees = []

        for ID, item in iter_dict.items():
            product.append(item.instrument)
            status.append(item.status)
            ids.append(item.id)
            size.append(item.size)
            direction.append(item.direction)
            times_list.append(item.order_time)
            order_type.append(item.order_type)
            order_price.append(item.order_price)
            stop_price.append(item.stop_loss)
            take_price.append(item.take_profit)
            reason.append(item.reason)

        if trades is not None:
            for trade_id, trade in iter_dict.items():
                entry_time.append(trade.time_filled)
                fill_price.append(trade.fill_price)
                profit.append(trade.profit)
                portfolio_balance.append(trade.balance)
                exit_times.append(trade.exit_time)
                exit_prices.append(trade.exit_price)
                fees.append(trade.fees)
                if trade.status == "closed":
                    if type(trade.exit_time) == str:
                        exit_dt = datetime.strptime(
                            trade.exit_time, "%Y-%m-%d %H:%M:%S%z"
                        )
                        entry_dt = datetime.strptime(
                            trade.time_filled, "%Y-%m-%d %H:%M:%S%z"
                        )
                        trade_duration.append(
                            exit_dt.timestamp() - entry_dt.timestamp()
                        )
                    elif isinstance(trade.exit_time, pd.Timestamp):
                        trade_duration.append(
                            (trade.exit_time - trade.time_filled).total_seconds()
                        )
                    elif trade.exit_time is None:
                        # Weird edge case
                        trade_duration.append(None)
                    else:
                        trade_duration.append(
                            trade.exit_time.timestamp() - trade.time_filled.timestamp()
                        )
                else:
                    trade_duration.append(None)

            dataframe = pd.DataFrame(
                {
                    "instrument": product,
                    "status": status,
                    "ID": ids,
                    "order_price": order_price,
                    "order_time": times_list,
                    "fill_time": entry_time,
                    "fill_price": fill_price,
                    "size": size,
                    "direction": direction,
                    "stop_loss": stop_price,
                    "take_profit": take_price,
                    "profit": profit,
                    "balance": portfolio_balance,
                    "exit_time": exit_times,
                    "exit_price": exit_prices,
                    "trade_duration": trade_duration,
                    "fees": fees,
                },
                index=pd.to_datetime(entry_time),
            )

            # Fill missing values for balance
            dataframe.balance.fillna(method="ffill", inplace=True)

        else:
            # Order summary
            dataframe = pd.DataFrame(
                {
                    "instrument": product,
                    "status": status,
                    "order_id": ids,
                    "order_type": order_type,
                    "order_price": order_price,
                    "order_time": times_list,
                    "size": size,
                    "direction": direction,
                    "stop_loss": stop_price,
                    "take_profit": take_price,
                    "reason": reason,
                },
                index=pd.to_datetime(times_list),
            )

        dataframe = dataframe.sort_index()

        # Add broker name column
        dataframe["broker"] = broker_name

        # Filter by instrument
        if instrument is not None:
            dataframe = dataframe[dataframe["instrument"] == instrument]

        return dataframe

    def summary(self) -> dict:
        """Constructs a trading summary for printing."""
        # Initialise trade results dict
        trade_results = {}

        # Analyse account history
        if len(self.account_history) > 0:
            # The account was open for some nonzero time period
            starting_balance = self.account_history["equity"].iloc[0]
            ending_balance = self.account_history["equity"].iloc[-1]
            ending_NAV = self.account_history["NAV"].iloc[-1]
            abs_return = ending_balance - starting_balance
            pc_return = 100 * abs_return / starting_balance
            floating_pnl = ending_NAV - ending_balance
            max_drawdown = min(self.account_history.drawdown)

            # Save results to dict
            trade_results["start"] = self.account_history.index[0]
            trade_results["end"] = self.account_history.index[-1]
            trade_results["starting_balance"] = starting_balance
            trade_results["ending_balance"] = ending_balance
            trade_results["ending_NAV"] = ending_NAV
            trade_results["abs_return"] = abs_return
            trade_results["pc_return"] = pc_return
            trade_results["floating_pnl"] = floating_pnl
            trade_results["max_drawdown"] = max_drawdown

        # All trades
        no_trades = len(self.trade_history)
        trade_results["no_trades"] = no_trades
        trade_results["no_long_trades"] = len(
            self.trade_history[self.trade_history["direction"] > 0]
        )
        trade_results["no_short_trades"] = len(
            self.trade_history[self.trade_history["direction"] < 0]
        )

        if no_trades > 0:
            # Initialise all_trades dict
            trade_results["all_trades"] = {}

            # Calculate positions still open
            # TODO - debug below with multi asset backtest
            try:
                trade_results["no_open"] = sum(self.position_history.iloc[-1] > 0)
            except:
                trade_results["no_open"] = 0

            # Analyse winning positions
            # wins = self.isolated_position_history[
            #     self.isolated_position_history.profit > 0
            # ]
            # avg_win = np.mean(wins.profit)
            # max_win = np.max(wins.profit)

            # Analyse losing positions
            # loss = self.isolated_position_history[
            #     self.isolated_position_history.profit < 0
            # ]
            # avg_loss = abs(np.mean(loss.profit))
            # max_loss = abs(np.min(loss.profit))

            # Performance
            # win_rate = 100 * len(wins) / no_trades
            # longest_win_streak, longest_lose_streak = get_streaks(
            #     self.isolated_position_history
            # )
            # try:
            #     avg_trade_duration = np.nanmean(
            #         self.isolated_position_history.trade_duration.values
            #     )
            #     trade_results["all_trades"]["avg_trade_duration"] = str(
            #         timedelta(seconds=int(avg_trade_duration))
            #     )
            # except TypeError:
            #     # Position has not been closed yet
            #     trade_results["all_trades"]["avg_trade_duration"] = None

            # Trade durations
            # duration_list = [
            #     i
            #     for i in self.isolated_position_history.trade_duration.values
            #     if i is not None
            # ]
            # if len(duration_list) > 0:
            #     min_trade_duration = np.nanmin(duration_list)
            #     max_trade_duration = np.nanmax(duration_list)
            # else:
            #     min_trade_duration = None
            #     max_trade_duration = None

            # Fees
            total_fees = self.trade_history.fee.sum()

            # Volume traded
            total_volume = (
                self.trade_history["size"] * self.trade_history["fill_price"]
            ).values.sum()

            # trade_results["all_trades"]["avg_win"] = avg_win
            # trade_results["all_trades"]["max_win"] = max_win
            # trade_results["all_trades"]["avg_loss"] = avg_loss
            # trade_results["all_trades"]["max_loss"] = max_loss
            # trade_results["all_trades"]["win_rate"] = win_rate
            # trade_results["all_trades"]["win_streak"] = longest_win_streak
            # trade_results["all_trades"]["lose_streak"] = longest_lose_streak
            trade_results["all_trades"]["avg_win"] = 0
            trade_results["all_trades"]["max_win"] = 0
            trade_results["all_trades"]["avg_loss"] = 0
            trade_results["all_trades"]["max_loss"] = 0
            trade_results["all_trades"]["win_rate"] = 0
            trade_results["all_trades"]["win_streak"] = 0
            trade_results["all_trades"]["lose_streak"] = 0
            trade_results["all_trades"]["total_volume"] = total_volume
            max_trade_duration = None
            min_trade_duration = None

            if max_trade_duration is not None:
                trade_results["all_trades"]["longest_trade"] = str(
                    timedelta(seconds=int(max_trade_duration))
                )
            else:
                trade_results["all_trades"]["longest_trade"] = str(None)

            if min_trade_duration is not None:
                trade_results["all_trades"]["shortest_trade"] = str(
                    timedelta(seconds=int(min_trade_duration))
                )
            else:
                trade_results["all_trades"]["shortest_trade"] = str(None)

            trade_results["all_trades"]["total_fees"] = total_fees

        # Cancelled orders
        trade_results["no_cancelled"] = len(self.cancelled_orders)

        # Long positions
        # long_positions = self.isolated_position_history[
        #     self.isolated_position_history["direction"] > 0
        # ]
        # no_long = len(long_positions)
        # trade_results["long_positions"] = {}
        # trade_results["long_positions"]["total"] = no_long
        # if no_long > 0:
        #     long_wins = long_positions[long_positions.profit > 0]
        #     avg_long_win = np.mean(long_wins.profit)
        #     max_long_win = np.max(long_wins.profit)
        #     long_loss = long_positions[long_positions.profit < 0]
        #     avg_long_loss = abs(np.mean(long_loss.profit))
        #     max_long_loss = abs(np.min(long_loss.profit))
        #     long_wr = 100 * len(long_positions[long_positions.profit > 0]) / no_long

        #     trade_results["long_positions"]["avg_long_win"] = avg_long_win
        #     trade_results["long_positions"]["max_long_win"] = max_long_win
        #     trade_results["long_positions"]["avg_long_loss"] = avg_long_loss
        #     trade_results["long_positions"]["max_long_loss"] = max_long_loss
        #     trade_results["long_positions"]["long_wr"] = long_wr

        # # Short positions
        # short_positions = self.isolated_position_history[
        #     self.isolated_position_history["direction"] < 0
        # ]
        # no_short = len(short_positions)
        # trade_results["short_positions"] = {}
        # trade_results["short_positions"]["total"] = no_short
        # if no_short > 0:
        #     short_wins = short_positions[short_positions.profit > 0]
        #     avg_short_win = np.mean(short_wins.profit)
        #     max_short_win = np.max(short_wins.profit)
        #     short_loss = short_positions[short_positions.profit < 0]
        #     avg_short_loss = abs(np.mean(short_loss.profit))
        #     max_short_loss = abs(np.min(short_loss.profit))
        #     short_wr = 100 * len(short_positions[short_positions.profit > 0]) / no_short

        #     trade_results["short_positions"]["avg_short_win"] = avg_short_win
        #     trade_results["short_positions"]["max_short_win"] = max_short_win
        #     trade_results["short_positions"]["avg_short_loss"] = avg_short_loss
        #     trade_results["short_positions"]["max_short_loss"] = max_short_loss
        #     trade_results["short_positions"]["short_wr"] = short_wr

        return trade_results


class AbstractDataStream(Broker):
    """Custom data feed base class. Wrapper around broker object without any
    private trading methods."""

    @abstractmethod
    def __init__(self, config: dict[str, any]) -> None:
        pass


class DataStream(AbstractDataStream):
    """Custom data feed base class. Wrapper around broker object without any
    private trading methods."""

    def __init__(self, config: dict[str, any]) -> None:
        self._data_broker = self

    @property
    def data_broker(self):
        return self._data_broker

    def __repr__(self):
        return "DataStreamer"

    def __str__(self):
        return "DataStreamer"


class LocalDataStream(DataStream):
    """Local data stream object."""

    def __init__(self, config: dict[str, any]) -> None:
        # Unpack parameters
        self._directory = config["directory"]
        self._data_dict: dict[str, str] = config["data_dict"]
        self._data_path_mapper = config["data_path_mapper"]

    def get_candles(
        self,
        instrument: str,
        granularity: str = None,
        count: int = None,
        start_time: datetime = None,
        end_time: datetime = None,
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        # TODO - test with portfolio

        # Get filepath
        if self._data_dict:
            # Local using filenames specified
            filepath = self._data_dict.get(instrument)

        elif self._data_path_mapper:
            # Use mapper function
            prefix = self._data_path_mapper(instrument)
            filepath = os.path.join(self._directory, f"{prefix}.csv")

        else:
            # Use instrument directory
            filepath = os.path.join(self._directory, f"{instrument}.csv")

        # Load
        candles = pd.read_csv(filepath, index_col=0, parse_dates=True)

        return candles

    def get_orderbook(self, instrument: str, *args, **kwargs):
        raise Exception("Orderbook data is not available from the local datastreamer.")

    def get_public_trades(self, instrument: str, *args, **kwargs):
        raise Exception(
            "Public trade data is not available from the local datastreamer."
        )


class CcxtDownloadStreamer(DataStream):
    """Data Streamer for CCXT Download."""

    def __init__(self, config: dict[str, any]):
        # Save CCXT-Download attributes
        self.data_directory = DEFAULT_DOWNLOAD_DIR
        self._cache_length = 3

        # Initialise cache
        self._cache: dict[str, pd.DataFrame] = {}

    def get_candles(
        self,
        instrument: str,
        granularity: str = None,
        start_time: datetime = None,
        end_time: datetime = None,
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        # TODO - need to check data ranges, same as in virtual broker?
        if instrument not in self._cache:
            # Load from data lake
            candles = load_data(
                exchange="bybit",  # TODO - need to get this from somewhere
                data_type=CANDLES,
                data_type_id=granularity,
                symbols=[instrument],
                start_date=start_time,
                end_date=end_time,
            )
            self._cache[instrument] = candles

        else:
            candles = self._cache[instrument]

        return candles

    def get_orderbook(self, instrument: str, *args, **kwargs):
        candles = self._cache[instrument]
        # Use local orderbook method.
        return candles

    def get_public_trades(self, instrument: str, *args, **kwargs):
        # TODO - implement
        return []


class TradeWatcher:
    """Watches trade snapshots to detect new trades."""

    def __init__(self) -> None:
        self.last_trade_time = None
        self.latest_trades = []

    def update(self, trades):
        """Updates the trades being monitored for change."""
        if trades[0]["time"] != self.last_trade_time:
            # Trade update
            self.last_trade_time = trades[0]["time"]

            for trade in trades:
                if trade["time"] != self.last_trade_time:
                    break
                self.latest_trades.append(trade)

    def get_latest_trades(self):
        """Returns the latest (unseen) trades."""
        latest_trades = self.latest_trades
        self.latest_trades = []
        return latest_trades


class Monitor:
    def __init__(
        self,
        config_filepath: Optional[str] = None,
        config: Optional[dict] = None,
        *args,
        **kwargs,
    ) -> None:
        """Construct a Monitor instance.

        Parameters
        ----------
        config_filepath : str, None
            The absolute filepath of the monitor yaml configuration file.
            The default is None.

        config : dict, optional
            The monitor configuration dictionary. The default is None.
        """
        # Initialise attributes
        self.port = None
        self.broker = None
        self.picklefile = None
        self.environment = None
        self.initial_nav = None
        self.max_nav = None
        self.sleep_time = None

        if config_filepath is not None:
            # Read monitor config
            print("Loading monitor configuration from file.")
            config = read_yaml(config_filepath)

        elif config is None and config_filepath is None:
            # Use kwargs as config
            config = kwargs

        # Overwrite config with kwargs
        for key, val in kwargs.items():
            if val is not None:
                config[key] = val

        # Check config keys
        required_keys = [
            "port",
            "environment",
            "initial_nav",
            "max_nav",
            "sleep_time",
        ]
        for key in required_keys:
            if key not in config.keys():
                print(f"Error: missing configuration key: '{key}'")
                sys.exit()

        # Check for broker key
        if "picklefile" in config and config["picklefile"] is not None:
            self.picklefile = config["picklefile"]
        else:
            try:
                self.broker = config["broker"]
            except KeyError:
                print("Please specify the broker name or pickle file path to monitor.")
                sys.exit()

        # Unpack config and assign attributes
        for key, val in config.items():
            setattr(self, key, val)

    def _initialise(self) -> None:
        """Initialise the monitor."""
        # Set up instrumentation
        self.nav_gauge = Gauge("nav_gauge", "Net Asset Value gauge.")
        self.drawdown_gauge = Gauge("drawdown_gauge", "Current drawdown gauge.")
        self.max_pos_gauge = Gauge(
            "max_position_gauge", "Maximum position notional gauge."
        )
        self.max_pos_frac_gauge = Gauge(
            "max_pos_frac_gauge", "Maximum position NAV fraction gauge."
        )
        self.abs_PnL_gauge = Gauge("abs_pnl_gauge", "Absolute ($) PnL gauge.")
        self.rel_PnL_gauge = Gauge("rel_pnl_gauge", "Relative (%) PnL gauge.")
        self.pos_gauge = Gauge("pos_gauge", "Number of open positions gauge.")
        self.total_exposure_gauge = Gauge(
            "total_exposure_gauge", "Total exposure gauge."
        )
        self.net_exposure_gauge = Gauge("net_exposure_gauge", "Total exposure gauge.")
        self.leverage_gauge = Gauge("leverage_gauge", "Total leverage gauge.")

        # Start up the server to expose the metrics
        try:
            Monitor.start_server(self.port)
        except OSError:
            print(f"Server on port {self.port} already in use. Terminating to restart.")

            # Kill existing server
            from psutil import process_iter
            from signal import SIGKILL  # or SIGTERM

            for proc in process_iter():
                for conns in proc.connections(kind="inet"):
                    if conns.laddr.port == self.port:
                        proc.send_signal(SIGKILL)  # or SIGKILL

            # Start server
            Monitor.start_server(self.port)

    def run(
        self,
    ) -> None:
        """Runs the monitor indefinitely."""
        # Initialise
        self._initialise()

        # Begin monitor loop
        broker = None
        print(f"Monitoring with {self.sleep_time} second updates.")
        deploy_time = datetime.now().timestamp()
        while True:
            try:
                # Get broker object
                broker = self.get_broker(broker)

                # Query broker
                nav = broker.get_NAV()
                if self.initial_nav is None:
                    self.initial_nav = nav
                if self.max_nav is None:
                    self.max_nav = nav

                positions = broker.get_positions()
                pnl = nav - self.initial_nav
                rel_pnl = pnl / self.initial_nav

                # Calculate drawdown
                if nav > self.max_nav:
                    self.max_nav = nav
                drawdown = -min(0, -(self.max_nav - nav) / self.max_nav)

                # Calculate total exposure
                total_exposure = 0
                net_exposure = 0
                max_pos_notional = 0
                pos_pnl = {}
                for instrument, position in positions.items():
                    pos_pnl[instrument] = position.pnl
                    pos_notional = abs(position.notional)
                    total_exposure += pos_notional
                    net_exposure += position.direction * position.notional

                    if pos_notional > max_pos_notional:
                        # Update maximum position notional value
                        max_pos_notional = pos_notional

                # Calculate max position fraction
                max_pos_frac = max_pos_notional / nav

                # Calculate leverage
                leverage = total_exposure / nav

                # Update metrics
                self.nav_gauge.set(nav)
                self.abs_PnL_gauge.set(pnl)
                self.rel_PnL_gauge.set(rel_pnl)
                self.pos_gauge.set(len(positions))
                self.total_exposure_gauge.set(total_exposure)
                self.net_exposure_gauge.set(net_exposure)
                self.leverage_gauge.set(leverage)
                self.drawdown_gauge.set(drawdown)
                self.max_pos_gauge.set(max_pos_notional)
                self.max_pos_frac_gauge.set(max_pos_frac)

                # Sleep
                time.sleep(
                    self.sleep_time - ((time.time() - deploy_time) % self.sleep_time)
                )

            except KeyboardInterrupt:
                print("\n\nStopping monitoring.")
                break

            except Exception as e:
                # Print exception
                print(e)

                # Also sleep briefly
                time.sleep(3)

                # Reconnect to broker
                broker = self.get_broker(None)

    @staticmethod
    def start_server(port):
        """Starts the http server for Prometheus."""
        start_http_server(port)
        print(f"Server started on port {port}.")

    def get_broker(self, broker) -> AbstractBroker:
        """Returns the broker object."""
        if self.picklefile is not None:
            # Load broker instance from pickle
            broker = unpickle_broker(self.picklefile)

        elif broker is not None:
            # Use existing broker instance
            pass

        else:
            # Create broker instance
            print(f"Connecting to {self.broker} ({self.environment} environment)...")
            at = autotrader.AutoTrader()
            at.configure(verbosity=0)
            at.configure(broker=self.broker, environment=self.environment, verbosity=0)
            broker = at.run()
            print("  Done.")
        return broker



import pandas as pd
from datetime import datetime
from abc import ABC, abstractmethod
from autotrader.brokers.broker import Broker
from autotrader.comms.notifier import Notifier
from typing import List, Union, TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from autotrader import Order


class Strategy(ABC):
    @abstractmethod
    def __init__(
        self,
        parameters: dict,
        instrument: str,
        broker: Broker,
        notifier: Notifier,
        logger_kwargs: dict[str, any],
        *args,
        **kwargs
    ) -> None:
        """Instantiate the strategy. This gets called from the AutoTraderBot assigned to
        this strategy.

        Parameters
        ----------
        parameters : dict
            The strategy parameters.

        instrument : str
            The instrument to trade.

        broker : Broker
            The broker connection.

        notifier : Notifier | None
            The notifier object. If notify is not set > 0, then this will be a NoneType object.

        logger_kwargs : dict[str, any]
            The logger configuration for this instance of AutoTrader. This can be unpacked
            into the get_logger() utility function when creating strategy loggers.
        """
        super().__init__()

    @abstractmethod
    def generate_signal(self, dt: datetime) -> Optional[Union["Order", List["Order"]]]:
        """Generate trading signals based on the data supplied."""

    def stop_trading():
        """Self destruct this instance of AutoTrader to stop any further trading."""
        pass

    def create_plotting_indicators(self, data: pd.DataFrame):
        """This method gets called with an entire backtest dataset to allow you to
        plot indicators."""
